/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <signal.h>
#include "spamdyke.h"
#include "tls.h"
#include "dns.h"
#include "environment.h"
#include "configuration.h"
#include "usage.h"
#include "search_fs.h"
#include "exec.h"
#include "base64.h"
#include "log.h"

struct rejection_data *rejection = REJECTION_DATA;

/*
 * Expects:
 *   strlen_haystack == strlen(haystack)
 *   return_address is a preallocated buffer
 *   max_address_len is the size of the return_address buffer
 *   strlen_username an int buffer if the length of the username is desired or NULL if not
 *
 * Return value:
 *   ERROR: 0
 *   SUCCESS: total length of the returned address
 */
int find_address(char *haystack, int strlen_haystack, char *return_address, int max_address_len, int *strlen_username)
  {
  int return_value;
  int i;
  int j;
  int haystack_start;
  int domain_start;
  int domain_end;
  int address_start;
  int address_end;
  int batv_address_end;

  return_value = 0;

  if ((return_address != NULL) &&
      (max_address_len > 0))
    {
    if ((haystack != NULL) &&
        (strlen_haystack > 0))
      {
      for (haystack_start = 0; (haystack_start < strlen_haystack) && (haystack[haystack_start] != ':'); haystack_start++);
      for (; (haystack_start < strlen_haystack) && !isalnum((int)haystack[haystack_start]); haystack_start++);

      domain_start = -1;
      address_start = -1;
      address_end = -1;
      batv_address_end = -1;

      for (j = haystack_start; (j < strlen_haystack) && !isspace((int)haystack[j]); j++);
      for (i = j - 1; (i >= haystack_start) && !isalnum((int)haystack[i]); i--);
      domain_end = i + 1;

      for (i = domain_end - 1; i >= haystack_start; i--)
        if (!isalnum((int)haystack[i]))
          {
          switch (haystack[i])
            {
            case '@':
              if (domain_start == -1)
                {
                domain_start = i + 1;
                address_end = i;
                continue;
                }

              break;
            case '/':
              if (address_end != -1)
                batv_address_end = i;

              continue;
              break;
            case '&':
            case ':':
            case '.':
            case '#':
            case '[':
            case ']':
            case '-':
            case '"':
            case '\\':
            case '_':
              continue;
              break;
            case '=':
              /* An equals sign followed by a slash indicates a BATV address. */
              if (batv_address_end != -1)
                address_end = batv_address_end;
              else
                continue;

              break;
            case ',':
              /* Commas indicate relay paths, which spamdyke ignores. */
            case '!':
              /* Exclamation points indicate bang paths, which spamdyke ignores. */
            default:
              break;
            }

          break;
          }

      address_start = i + 1;
      if (domain_start == -1)
        {
        address_end = domain_end;
        domain_end = -1;

        return_value = snprintf(return_address, max_address_len, "%.*s", address_end - address_start, haystack + address_start);
        }
      else
        return_value = snprintf(return_address, max_address_len, "%.*s@%.*s", address_end - address_start, haystack + address_start, domain_end - domain_start, haystack + domain_start);

      if (strlen_username != NULL)
        *strlen_username = address_end - address_start;
      }
    else
      return_address[0] = '\0';
    }

  return(return_value);
  }

/*
 * Expects:
 *   action ==
 *     LOG_ACTION_TLS_PASSTHROUGH_START - insert log header showing TLS passthrough has begun, no data
 *     LOG_ACTION_AUTH_FAILURE - insert log header showing SMTP AUTH was attempted but failed, no data
 *     LOG_ACTION_AUTH_SUCCESS - insert log header showing SMTP AUTH was successful, username is stored in current_settings->smtp_auth_username, no data
 *     LOG_ACTION_TLS_START - insert log header showing TLS session has begun, no data
 *     LOG_ACTION_TLS_END - insert log header showing TLS session has ended, no data
 *     LOG_ACTION_NONE - no header, do not log
 *     LOG_ACTION_REMOTE_FROM - no header, log data from remote to child
 *     LOG_ACTION_CHILD_FROM - no header, log data from child to remote
 *     LOG_ACTION_CHILD_FROM_DISCARDED - no header, log discarded data from child to remote
 *     LOG_ACTION_FILTER_FROM - no header, log data injected by spamdyke to remote
 *     LOG_ACTION_FILTER_TO - no header, log data injected by spamdyke to child
 *   target_fd == file descriptor destination for data or -1 if data should not be output
 *
 * Return value:
 *   ERROR: -1
 *   SUCCESS: bytes output to target_fd
 */
int output_writeln(struct filter_settings *current_settings, int action, int target_fd, char *data, int data_length)
  {
  static FILE *log_file = NULL;
  static int last_action = -1;
  static char *prefixes[] = LOG_ACTION_PREFIX;
  static char last_char = '\0';
  int return_value;
  int i;
  char log_filename[MAX_BUF + 1];
  struct tm *tmp_tm;
  time_t tmp_time;
  char *data_ptr;

  return_value = 0;

  if (action != LOG_ACTION_NONE)
    {
    tmp_tm = NULL;

    if ((current_settings == NULL) &&
        (log_file != NULL))
      {
      tmp_time = time(NULL);
      tmp_tm = localtime(&tmp_time);

      fprintf(log_file, "\n%.2d/%.2d/%d %.2d:%.2d:%.2d CLOSED\n", tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

      fclose(log_file);
      log_file = NULL;
      }
    else if ((current_settings != NULL) &&
             (current_settings->log_dir != NULL) &&
             (log_file == NULL))
      {
      tmp_time = time(NULL);
      tmp_tm = localtime(&tmp_time);

      if (current_settings->strlen_server_name > 0)
        snprintf(log_filename, MAX_BUF, "%s/%d%.2d%.2d_%.2d%.2d%.2d_%s_%s", current_settings->log_dir, tmp_tm->tm_year + 1900, tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, current_settings->server_ip, current_settings->server_name);
      else
        snprintf(log_filename, MAX_BUF, "%s/%d%.2d%.2d_%.2d%.2d%.2d_%s", current_settings->log_dir, tmp_tm->tm_year + 1900, tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, current_settings->server_ip);

      if ((log_file = fopen(log_filename, "w")) != NULL)
        {
        chmod(log_filename, CHMOD_MODE);
        fprintf(log_file, "%.2d/%.2d/%d %.2d:%.2d:%.2d STARTED: VERSION = %s, PID = " FORMAT_PID_T "\n", tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, VERSION_STRING, getpid());
        fprintf(log_file, "%.2d/%.2d/%d %.2d:%.2d:%.2d LEGEND: To remote host = %s; to child process = %s; blocked by filter = %s\n", tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, prefixes[LOG_ACTION_CHILD_FROM], prefixes[LOG_ACTION_REMOTE_FROM], prefixes[LOG_ACTION_CHILD_FROM_DISCARDED]);
        fprintf(log_file, "%.2d/%.2d/%d %.2d:%.2d:%.2d LEGEND: From filter to remote host = %s; from filter to child process = %s\n", tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, prefixes[LOG_ACTION_FILTER_FROM], prefixes[LOG_ACTION_FILTER_TO]);
        }
      else
        {
        SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_OPEN_LOG "%s: %s", log_filename, strerror(errno));
        current_settings->log_dir = NULL;
        }
      }

    if (log_file != NULL)
      {
      if (tmp_tm == NULL)
        {
        tmp_time = time(NULL);
        tmp_tm = localtime(&tmp_time);
        }

      switch (action)
        {
        case LOG_ACTION_TLS_PASSTHROUGH_START:
          fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[LOG_ACTION_FILTER_FROM], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

          if (current_settings->smtp_auth_username[0] != '\0')
            fprintf(log_file, LOG_ACTION_PREFIX_AUTH "%s", current_settings->smtp_auth_username);

          fprintf(log_file, " - %s\n", LOG_MESSAGE_TLS_PASSTHROUGH_START);
          break;
        case LOG_ACTION_TLS_START:
          fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[LOG_ACTION_FILTER_FROM], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

          if (current_settings->smtp_auth_username[0] != '\0')
            fprintf(log_file, LOG_ACTION_PREFIX_AUTH "%s", current_settings->smtp_auth_username);

          fprintf(log_file, " - %s\n", LOG_MESSAGE_TLS_START);
          break;
        case LOG_ACTION_TLS_END:
          fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[LOG_ACTION_FILTER_FROM], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

          if (current_settings->smtp_auth_username[0] != '\0')
            fprintf(log_file, LOG_ACTION_PREFIX_AUTH "%s", current_settings->smtp_auth_username);

          fprintf(log_file, " - %s\n", LOG_MESSAGE_TLS_END);
          break;
        case LOG_ACTION_AUTH_SUCCESS:
          fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[LOG_ACTION_FILTER_FROM], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

          if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
            fprintf(log_file, LOG_ACTION_PREFIX_TLS_SPAMDYKE);
          else if (current_settings->tls_state == TLS_STATE_ACTIVE_PASSTHROUGH)
            fprintf(log_file, LOG_ACTION_PREFIX_TLS_PASSTHROUGH);

          if (current_settings->smtp_auth_username[0] != '\0')
            fprintf(log_file, LOG_ACTION_PREFIX_AUTH "%s", (current_settings->smtp_auth_username[0] != '\0') ? current_settings->smtp_auth_username : LOG_MISSING_DATA);

          fprintf(log_file, " - %s\n", LOG_MESSAGE_AUTH_SUCCESS);
          break;
        case LOG_ACTION_AUTH_FAILURE:
          fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[LOG_ACTION_FILTER_FROM], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

          if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
            fprintf(log_file, LOG_ACTION_PREFIX_TLS_SPAMDYKE);
          else if (current_settings->tls_state == TLS_STATE_ACTIVE_PASSTHROUGH)
            fprintf(log_file, LOG_ACTION_PREFIX_TLS_PASSTHROUGH);

          fprintf(log_file, " - %s%s\n", LOG_MESSAGE_AUTH_FAILURE, data);
          break;
        default:
          if (action != last_action)
            {
            fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[action], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

            if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
              fprintf(log_file, LOG_ACTION_PREFIX_TLS_SPAMDYKE);
            else if (current_settings->tls_state == TLS_STATE_ACTIVE_PASSTHROUGH)
              fprintf(log_file, LOG_ACTION_PREFIX_TLS_PASSTHROUGH);

            if (current_settings->smtp_auth_username[0] != '\0')
              fprintf(log_file, LOG_ACTION_PREFIX_AUTH "%s", current_settings->smtp_auth_username);

            fprintf(log_file, "\n");
            }
          break;
        }

      last_action = action;

      if (data_length >= 0)
        {
        if (current_settings->tls_state == TLS_STATE_ACTIVE_PASSTHROUGH)
          {
          int i;

          for (i = 0; i < data_length; i++)
            fprintf(log_file, "%.2x ", data[i]);
          }
        else
          fprintf(log_file, "%.*s", data_length, data);
        }

      fflush(NULL);
      }
    }

  if ((target_fd >= 0) &&
      (data_length > 0))
    {
    if ((action == LOG_ACTION_CHILD_FROM) || (action == LOG_ACTION_FILTER_FROM))
      {
      if ((return_value = NETWORK_WRITE(current_settings, target_fd, data, data_length)) == -1)
        SPAMDYKE_LOG_DEBUG(current_settings, LOG_ERROR_WRITE "%s", data_length, target_fd, strerror(errno));
      }
    else if (current_settings->tls_state == TLS_STATE_ACTIVE_PASSTHROUGH)
      {
      if ((return_value = write(target_fd, data, data_length)) == -1)
        SPAMDYKE_LOG_DEBUG(current_settings, LOG_ERROR_WRITE "%s", data_length, target_fd, strerror(errno));
      }
    else
      {
      data_ptr = data;
      while ((data_ptr - data) < data_length)
        {
        for (i = 0; (((data_ptr - data) + i) < data_length) && (data_ptr[i] != CHAR_LF); i++);
        if (((data_ptr - data) + i) < data_length)
          {
          if ((i > 0) &&
              (data_ptr[i - 1] == CHAR_CR))
            {
            if ((return_value = write(target_fd, data_ptr, i + 1)) == -1)
              {
              SPAMDYKE_LOG_DEBUG(current_settings, LOG_ERROR_WRITE "%s", i + 1, target_fd, strerror(errno));
              break;
              }
            }
          else if ((i > 0) ||
                   (last_char != CHAR_CR))
            {
            if ((i > 0) &&
                ((return_value = write(target_fd, data_ptr, i)) == -1))
              {
              SPAMDYKE_LOG_DEBUG(current_settings, LOG_ERROR_WRITE "%s", i, target_fd, strerror(errno));
              break;
              }

            if ((return_value = write(target_fd, STR_CRLF, STRLEN(STR_CRLF))) == -1)
              {
              SPAMDYKE_LOG_DEBUG(current_settings, LOG_ERROR_WRITE "%s", STRLEN(STR_CRLF), target_fd, strerror(errno));
              break;
              }
            }
          else
            if ((return_value = write(target_fd, data_ptr, 1)) == -1)
              {
              SPAMDYKE_LOG_DEBUG(current_settings, LOG_ERROR_WRITE "%s", 1, target_fd, strerror(errno));
              break;
              }

          data_ptr += i + 1;
          }
        else
          {
          if ((return_value = write(target_fd, data_ptr, i)) == -1)
            SPAMDYKE_LOG_DEBUG(current_settings, LOG_ERROR_WRITE "%s", i, target_fd, strerror(errno));

          break;
          }
        }

      last_char = data[data_length - 1];
      }
    }
  else
    {
    return_value = 0;
    last_char = '\0';
    }

  return(return_value);
  }

/*
 * Expects:
 *   target_fd == file descriptor destination or -1 for no output
 *   substitute_reject_severity == SMTP status code or NULL to use status from target_rejection
 *
 * Return value:
 *   ERROR: -1
 *   SUCCESS: bytes output to target_fd
 */
int output_write_rejection(struct filter_settings *current_settings, struct rejection_data *target_rejection, int target_fd, char *substitute_reject_severity)
  {
  int return_value;
  int strlen_buf;
  char output_buf[MAX_BUF + 1];

  if (current_settings->policy_location != NULL)
    strlen_buf = snprintf(output_buf, MAX_BUF, "%.*s%.*s%.*s%.*s%.*s", STRLEN_REJECT_SEVERITY, (substitute_reject_severity == NULL) ? target_rejection->reject_severity : substitute_reject_severity, target_rejection->strlen_reject_message, target_rejection->reject_message, (int)STRLEN(ERROR_URL), ERROR_URL, current_settings->strlen_policy_location, current_settings->policy_location, (int)STRLEN(REJECT_CRLF), REJECT_CRLF);
  else
    strlen_buf = snprintf(output_buf, MAX_BUF, "%.*s%.*s%.*s", STRLEN_REJECT_SEVERITY, (substitute_reject_severity == NULL) ? target_rejection->reject_severity : substitute_reject_severity, target_rejection->strlen_reject_message, target_rejection->reject_message, (int)STRLEN(REJECT_CRLF), REJECT_CRLF);

  return_value = output_writeln(current_settings, LOG_ACTION_FILTER_FROM, target_fd, output_buf, strlen_buf);

  fflush(NULL);

  return(return_value);
  }

/*
 * Expects:
 *   inbound_fd == inbound file descriptor from remote server
 *   outbound_fd == outbound file descriptor to remote server
 *
 * Return value:
 *   bitwise OR of FILTER_FLAG values
 */
int smtp_filter(int inbound_fd, int outbound_fd, char *input_line, int strlen_input_line, struct filter_settings *current_settings, char *envp[])
  {

#ifdef HAVE_LIBSSL
  static struct rejection_data tls_success = SUCCESS_TLS;
#endif /* HAVE_LIBSSL */

  static char *environment_hostname[] = ENVIRONMENT_HOSTNAME;
  static int strlen_environment_hostname[] = STRLEN_ENVIRONMENT_HOSTNAME;
  static struct rejection_data tls_failure = FAILURE_TLS;
  static struct rejection_data smtp_auth_success = SUCCESS_AUTH;
  static char *smtp_auth_types[] = SMTP_AUTH_TYPES;
  static int strlen_smtp_auth_types[] = STRLEN_SMTP_AUTH_TYPES;
  int return_value;
  int i;
  int j;
  int search_return;
  char target_address[MAX_BUF + 1];
  int strlen_target_address;
  int strlen_username;
  struct stat tmp_stat;
  FILE *tmp_file;
  char log_entry[MAX_BUF + 1];
  int continue_processing;
  char tmp_recipient_domain[MAX_BUF + 1];
  char tmp_recipient_username[MAX_BUF + 1];
  char tmp_sender_address[MAX_BUF + 1];
  int search_results;
  char decoded_content[MAX_BUF + 1];
  char *content_ptr;
  int strlen_content;
  char username[MAX_BUF + 1];
  char password[MAX_BUF + 1];
  int local_recipient;
  int graylist_index;

  return_value = FILTER_FLAG_PASS;

  if (current_settings->inside_data)
    {
    if ((strlen_input_line == STRLEN(SMTP_DATA_END)) &&
        (strncasecmp(SMTP_DATA_END, input_line, STRLEN(SMTP_DATA_END)) == 0))
      current_settings->inside_data = 0;
    else
      return_value = FILTER_FLAG_PASS | FILTER_FLAG_CHILD_RESPONSE_NOT_NEEDED;
    }
  else if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
           (current_settings->smtp_auth_state == SMTP_AUTH_STATE_CHALLENGE_1_SENT))
    {
    current_settings->smtp_auth_state = SMTP_AUTH_STATE_RESPONSE_1_SEEN;

    switch (current_settings->smtp_auth_type)
      {
      case SMTP_AUTH_PLAIN:
        strlen_content = base64_decode((unsigned char *)decoded_content, MAX_BUF, (unsigned char *)input_line, strlen_input_line);
        decoded_content[strlen_content] = '\0';

        for (i = 0; (i < strlen_content) && (decoded_content[i] != '\0'); i++);
        content_ptr = decoded_content + i;
        strlen_content -= i;

        if (current_settings->smtp_auth_command != NULL)
          {
          if ((strlen_content >= 3) &&
              (content_ptr[0] == '\0'))
            {
            snprintf(username, MAX_BUF, "%s", content_ptr + 1);
            for (i = 1; ((i < strlen_content) && (content_ptr[i] != '\0')); i++);
            if (i < strlen_content)
              {
              snprintf(password, MAX_BUF, "%s", content_ptr + i + 1);

              for (i = 0; current_settings->smtp_auth_command[i] != NULL; i++)
                if (exec_checkpassword(current_settings, current_settings->smtp_auth_command[i], envp, username, password, NULL))
                  {
                  if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
                    {
                    output_write_rejection(current_settings, &smtp_auth_success, outbound_fd, NULL);
                    return_value = FILTER_FLAG_INTERCEPT;
                    }
                  else
                    return_value = FILTER_FLAG_PASS;

                  snprintf(current_settings->smtp_auth_username, MAX_BUF, "%s", username);
                  current_settings->filter_action = FILTER_DECISION_DO_NOT_FILTER;
                  current_settings->rejection = NULL;

                  output_writeln(current_settings, LOG_ACTION_AUTH_SUCCESS, -1, NULL, -1);
                  break;
                  }
              }
            }

          if (current_settings->filter_action == FILTER_DECISION_DO_FILTER)
            {
            output_writeln(current_settings, LOG_ACTION_AUTH_FAILURE, -1, username, -1);

            if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
              {
              output_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
              return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_CONTINUE;
              }
            else
              return_value = FILTER_FLAG_PASS;
            }

          current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
          }
        else
          {
          if ((strlen_content >= 3) &&
              (content_ptr[0] == '\0'))
            snprintf(current_settings->smtp_auth_response, MAX_BUF, "%s", content_ptr + 1);

          current_settings->smtp_auth_state = SMTP_AUTH_STATE_UNKNOWN;
          return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
          }

        break;
      case SMTP_AUTH_LOGIN:
        strncpy(current_settings->smtp_auth_response, input_line, MINVAL(MAX_BUF, strlen_input_line));
        current_settings->smtp_auth_response[MINVAL(MAX_BUF, strlen_input_line)] = '\0';

        if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
          {
          strlen_content = snprintf(current_settings->smtp_auth_challenge, MAX_BUF, "%s%s%s", REJECT_SEVERITY_AUTH_CHALLENGE, SMTP_AUTH_LOGIN_CHALLENGE_2, REJECT_CRLF);
          output_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->smtp_auth_challenge, strlen_content);
          current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_2_SENT;

          return_value = FILTER_FLAG_INTERCEPT;
          }
        else
          return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;

        break;
      case SMTP_AUTH_CRAM_MD5:
        strlen_content = base64_decode((unsigned char *)decoded_content, MAX_BUF, (unsigned char *)input_line, strlen_input_line);
        decoded_content[strlen_content] = '\0';

        if (current_settings->smtp_auth_command != NULL)
          {
          if (strlen_content > 0)
            {
            for (i = 0; (i < strlen_content) && (decoded_content[i] != ' '); i++);

            if (i < strlen_content)
              {
              snprintf(username, MAX_BUF, "%.*s", i, decoded_content);
              snprintf(password, MAX_BUF, "%.*s", (strlen_content - i) - 1, decoded_content + i + 1);

              strlen_content = strlen(current_settings->smtp_auth_challenge);
              for (i = STRLEN(SMTP_AUTH_CHALLENGE); (i < strlen_content) && (current_settings->smtp_auth_challenge[i] != ' '); i++);
              if (i < strlen_content)
                {
                strlen_content = base64_decode((unsigned char *)decoded_content, MAX_BUF, (unsigned char *)current_settings->smtp_auth_challenge + i, strlen_content - i);
                decoded_content[strlen_content] = '\0';

                for (i = 0; current_settings->smtp_auth_command[i] != NULL; i++)
                  if (exec_checkpassword(current_settings, current_settings->smtp_auth_command[i], envp, username, password, decoded_content))
                    {
                    if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
                      {
                      output_write_rejection(current_settings, &smtp_auth_success, outbound_fd, NULL);
                      return_value = FILTER_FLAG_INTERCEPT;
                      }
                    else
                      return_value = FILTER_FLAG_PASS;

                    snprintf(current_settings->smtp_auth_username, MAX_BUF, "%s", username);
                    current_settings->filter_action = FILTER_DECISION_DO_NOT_FILTER;
                    current_settings->rejection = NULL;

                    output_writeln(current_settings, LOG_ACTION_AUTH_SUCCESS, -1, NULL, -1);
                    break;
                    }
                }
              }
            }

          if (current_settings->filter_action == FILTER_DECISION_DO_FILTER)
            {
            output_writeln(current_settings, LOG_ACTION_AUTH_FAILURE, -1, username, -1);

            if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
              {
              output_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
              return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_CONTINUE;
              }
            else
              return_value = FILTER_FLAG_PASS;
            }

          current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
          }
        else
          {
          if (strlen_content > 0)
            {
            for (i = 0; (i < strlen_content) && (decoded_content[i] != ' '); i++);

            if (i < strlen_content)
              snprintf(current_settings->smtp_auth_response, MAX_BUF, "%.*s", i, decoded_content);
            }

          current_settings->smtp_auth_state = SMTP_AUTH_STATE_UNKNOWN;
          return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
          }

        break;
      default:
        /* The client is using a protocol spamdyke doesn't support. Continue listening for the server's response and trust it. */
        current_settings->smtp_auth_state = SMTP_AUTH_STATE_UNKNOWN;
        return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
        break;
      }
    }
  else if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
           (current_settings->smtp_auth_state == SMTP_AUTH_STATE_CHALLENGE_2_SENT))
    {
    switch (current_settings->smtp_auth_type)
      {
      case SMTP_AUTH_LOGIN:
        strlen_content = base64_decode((unsigned char *)username, MAX_BUF, (unsigned char *)current_settings->smtp_auth_response, strlen(current_settings->smtp_auth_response));
        username[strlen_content] = '\0';

        for (j = 0; (j < strlen_input_line) && !isalnum((int)input_line[j]); j++);
        strlen_content = base64_decode((unsigned char *)password, MAX_BUF, (unsigned char *)(input_line + j), strlen_input_line - j);
        password[strlen_content] = '\0';

        if (current_settings->smtp_auth_command != NULL)
          {
          for (i = 0; current_settings->smtp_auth_command[i] != NULL; i++)
            if (exec_checkpassword(current_settings, current_settings->smtp_auth_command[i], envp, username, password, NULL))
              {
              if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
                {
                output_write_rejection(current_settings, &smtp_auth_success, outbound_fd, NULL);
                return_value = FILTER_FLAG_INTERCEPT;
                }
              else
                return_value = FILTER_FLAG_PASS;

              snprintf(current_settings->smtp_auth_username, MAX_BUF, "%s", username);
              current_settings->filter_action = FILTER_DECISION_DO_NOT_FILTER;
              current_settings->rejection = NULL;

              output_writeln(current_settings, LOG_ACTION_AUTH_SUCCESS, -1, NULL, -1);
              break;
              }

          if (current_settings->filter_action == FILTER_DECISION_DO_FILTER)
            {
            output_writeln(current_settings, LOG_ACTION_AUTH_FAILURE, -1, username, -1);

            if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
              {
              output_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
              return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_CONTINUE;
              }
            else
              return_value = FILTER_FLAG_PASS;
            }

          current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
          }
        else
          {
          snprintf(current_settings->smtp_auth_response, MAX_BUF, "%s", username);
          current_settings->smtp_auth_state = SMTP_AUTH_STATE_UNKNOWN;
          return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
          }

        break;
      default:
        /* The client is using a protocol spamdyke doesn't support. Continue listening for the server's response and trust it. */
        current_settings->smtp_auth_state = SMTP_AUTH_STATE_UNKNOWN;
        return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
        break;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_HELO)) &&
           (strncasecmp(SMTP_HELO, input_line, STRLEN(SMTP_HELO)) == 0))
    {
    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->filter_grace <= FILTER_GRACE_NONE) &&
        (current_settings->rejection != NULL))
      {
      output_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_NONE);
      current_settings->filter_grace = FILTER_GRACE_EXPIRED;
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_EHLO)) &&
           (strncasecmp(SMTP_EHLO, input_line, STRLEN(SMTP_EHLO)) == 0))
    {
    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->smtp_auth_command != NULL))
      {
      return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_ADD;

#ifdef HAVE_LIBSSL
      if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
        return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_ADD | FILTER_FLAG_TLS_REMOVE;
      else if (current_settings->tls_certificate_file != NULL)
        return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_ADD | FILTER_FLAG_TLS_ADD;
      }
    else if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
      return_value = FILTER_FLAG_PASS | FILTER_FLAG_TLS_REMOVE;
    else if (current_settings->tls_certificate_file != NULL)
      {
      return_value = FILTER_FLAG_PASS | FILTER_FLAG_TLS_ADD;
#endif /* HAVE_LIBSSL */

      }
    else if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
             (current_settings->filter_grace <= FILTER_GRACE_NONE) &&
             (current_settings->rejection != NULL))
      {
      output_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_NONE);
      current_settings->filter_grace = FILTER_GRACE_EXPIRED;
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    }
  else if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
           (strlen_input_line >= STRLEN(SMTP_AUTH)) &&
           (strncasecmp(SMTP_AUTH, input_line, STRLEN(SMTP_AUTH)) == 0))
    {
    for (j = STRLEN(SMTP_AUTH); (j < strlen_input_line) && isspace((int)input_line[j]); j++);
    if (j < strlen_input_line)
      for (i = 0; smtp_auth_types[i] != NULL; i++)
        if ((strlen_input_line >= (strlen_smtp_auth_types[i] + j)) &&
            (strncasecmp(input_line + j, smtp_auth_types[i], strlen_smtp_auth_types[i]) == 0))
          {
          current_settings->smtp_auth_type = i;
          break;
          }

    current_settings->smtp_auth_state = SMTP_AUTH_STATE_CMD_SEEN;

    if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
      {
      switch (current_settings->smtp_auth_type)
        {
        case SMTP_AUTH_PLAIN:
          //FIXME: This code appears 3 times for processing AUTH PLAIN.  Can it be moved to a function?
          for (i = j + strlen_smtp_auth_types[SMTP_AUTH_PLAIN]; (i < strlen_input_line) && !isalnum((int)input_line[i]); i++);
          if ((i < strlen_input_line) &&
              isalnum((int)input_line[i]))
            {
            strlen_content = base64_decode((unsigned char *)decoded_content, MAX_BUF, (unsigned char *)(input_line + i), strlen_input_line - i);
            decoded_content[strlen_content] = '\0';

            for (i = 0; (i < strlen_content) && (decoded_content[i] != '\0'); i++);
            content_ptr = decoded_content + i;
            strlen_content -= i;

            if ((strlen_content >= 3) &&
                (content_ptr[0] == '\0'))
              {
              snprintf(username, MAX_BUF, "%s", content_ptr + 1);
              for (i = 1; ((i < strlen_content) && (content_ptr[i] != '\0')); i++);
              if (i < strlen_content)
                {
                snprintf(password, MAX_BUF, "%s", content_ptr + i + 1);

                for (i = 0; current_settings->smtp_auth_command[i] != NULL; i++)
                  if (exec_checkpassword(current_settings, current_settings->smtp_auth_command[i], envp, username, password, NULL))
                    {
                    if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
                      {
                      output_write_rejection(current_settings, &smtp_auth_success, outbound_fd, NULL);
                      return_value = FILTER_FLAG_INTERCEPT;
                      }
                    else
                      return_value = FILTER_FLAG_PASS;

                    snprintf(current_settings->smtp_auth_username, MAX_BUF, "%s", username);
                    current_settings->filter_action = FILTER_DECISION_DO_NOT_FILTER;
                    current_settings->rejection = NULL;

                    output_writeln(current_settings, LOG_ACTION_AUTH_SUCCESS, -1, NULL, -1);
                    break;
                    }
                }
              }

            if (current_settings->filter_action == FILTER_DECISION_DO_FILTER)
              {
              output_writeln(current_settings, LOG_ACTION_AUTH_FAILURE, -1, username, -1);

              if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
                {
                output_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
                return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_CONTINUE;
                }
              else
                return_value = FILTER_FLAG_PASS;
              }

            current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
            }
          else
            {
            strlen_content = snprintf(current_settings->smtp_auth_challenge, MAX_BUF, "%s%s", REJECT_SEVERITY_AUTH_CHALLENGE, REJECT_CRLF);
            output_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->smtp_auth_challenge, strlen_content);
            current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_1_SENT;
            }

          break;
        case SMTP_AUTH_LOGIN:
          for (i = j + strlen_smtp_auth_types[SMTP_AUTH_LOGIN]; (i < strlen_input_line) && !isalnum((int)input_line[i]); i++);
          if ((i < strlen_input_line) &&
              isalnum((int)input_line[i]))
            {
            strncpy(current_settings->smtp_auth_response, input_line + i, MINVAL(MAX_BUF, strlen_input_line - i));
            current_settings->smtp_auth_response[MINVAL(MAX_BUF, strlen_input_line - i)] = '\0';

            strlen_content = snprintf(current_settings->smtp_auth_challenge, MAX_BUF, "%s%s%s", REJECT_SEVERITY_AUTH_CHALLENGE, SMTP_AUTH_LOGIN_CHALLENGE_2, REJECT_CRLF);
            output_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->smtp_auth_challenge, strlen_content);
            current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_2_SENT;
            }
          else
            {
            strlen_content = snprintf(current_settings->smtp_auth_challenge, MAX_BUF, "%s%s%s", REJECT_SEVERITY_AUTH_CHALLENGE, SMTP_AUTH_LOGIN_CHALLENGE_1, REJECT_CRLF);
            output_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->smtp_auth_challenge, strlen_content);
            current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_1_SENT;
            }

          break;
        case SMTP_AUTH_CRAM_MD5:
          if ((current_settings->local_server_name == NULL) &&
              (current_settings->local_server_name_file != NULL))
            read_file(current_settings, current_settings->local_server_name_file, &current_settings->local_server_name, -1);

          if ((current_settings->local_server_name == NULL) &&
              (current_settings->local_server_name_command != NULL))
            exec_command(current_settings, current_settings->local_server_name_command, envp, NULL, &current_settings->local_server_name, -1);

          if (current_settings->local_server_name == NULL)
            for (i = 0; environment_hostname[i] != NULL; i++)
              if ((current_settings->local_server_name = find_environment_variable(envp, environment_hostname[i], strlen_environment_hostname[i])) != NULL)
                break;

          strlen_content = snprintf(decoded_content, MAX_BUF, "<%ld.%ld@%s>", random(), (long)time(NULL), (current_settings->local_server_name != NULL) ? current_settings->local_server_name : MISSING_LOCAL_SERVER_NAME);
          snprintf(current_settings->smtp_auth_challenge, MAX_BUF - STRLEN(REJECT_CRLF), "%s ", SMTP_AUTH_CHALLENGE);

          strlen_content = base64_encode((unsigned char *)(current_settings->smtp_auth_challenge + STRLEN(SMTP_AUTH_CHALLENGE) + 1), MAX_BUF - (STRLEN(SMTP_AUTH_CHALLENGE) + STRLEN(REJECT_CRLF) + 1), (unsigned char *)decoded_content, strlen_content) + STRLEN(SMTP_AUTH_CHALLENGE) + 1;
          strncpy(current_settings->smtp_auth_challenge + strlen_content, REJECT_CRLF, STRLEN(REJECT_CRLF));
          current_settings->smtp_auth_challenge[strlen_content + STRLEN(REJECT_CRLF)] = '\0';

          output_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->smtp_auth_challenge, strlen_content + STRLEN(REJECT_CRLF));

          current_settings->smtp_auth_challenge[strlen_content] = '\0';

          current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_1_SENT;

          break;
        default:
          /* The client requested an algorithm spamdyke didn't advertise. */
          output_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_UNKNOWN], outbound_fd, NULL);
          current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
          current_settings->filter_action = FILTER_DECISION_DO_FILTER;
          break;
        }

      return_value = FILTER_FLAG_INTERCEPT;
      }
    else
      switch (current_settings->smtp_auth_type)
        {
        case SMTP_AUTH_PLAIN:
          for (i = j + strlen_smtp_auth_types[SMTP_AUTH_PLAIN]; (i < strlen_input_line) && !isalnum((int)input_line[i]); i++);
          if ((i < strlen_input_line) &&
              isalnum((int)input_line[i]))
            {
            strlen_content = base64_decode((unsigned char *)decoded_content, MAX_BUF, (unsigned char *)(input_line + i), strlen_input_line - i);
            decoded_content[strlen_content] = '\0';

            for (i = 0; (i < strlen_content) && (decoded_content[i] != '\0'); i++);
            content_ptr = decoded_content + i;
            strlen_content -= i;

            if (current_settings->smtp_auth_command != NULL)
              {
              if ((strlen_content >= 3) &&
                  (content_ptr[0] == '\0'))
                {
                snprintf(username, MAX_BUF, "%s", content_ptr + 1);
                for (i = 1; ((i < strlen_content) && (content_ptr[i] != '\0')); i++);
                if (i < strlen_content)
                  {
                  snprintf(password, MAX_BUF, "%s", content_ptr + i + 1);

                  for (i = 0; current_settings->smtp_auth_command[i] != NULL; i++)
                    if (exec_checkpassword(current_settings, current_settings->smtp_auth_command[i], envp, username, password, NULL))
                      {
                      if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
                        {
                        output_write_rejection(current_settings, &smtp_auth_success, outbound_fd, NULL);
                        return_value = FILTER_FLAG_INTERCEPT;
                        }
                      else
                        return_value = FILTER_FLAG_PASS;

                      snprintf(current_settings->smtp_auth_username, MAX_BUF, "%s", username);
                      current_settings->filter_action = FILTER_DECISION_DO_NOT_FILTER;
                      current_settings->rejection = NULL;

                      output_writeln(current_settings, LOG_ACTION_AUTH_SUCCESS, -1, NULL, -1);
                      break;
                      }
                  }
                }

              if (current_settings->filter_action == FILTER_DECISION_DO_FILTER)
                {
                output_writeln(current_settings, LOG_ACTION_AUTH_FAILURE, -1, username, -1);

                if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
                  {
                  output_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
                  return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_CONTINUE;
                  }
                else
                  return_value = FILTER_FLAG_PASS;
                }

              current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
              }
            else
              {
              if ((strlen_content >= 3) &&
                  (content_ptr[0] == '\0'))
                snprintf(current_settings->smtp_auth_response, MAX_BUF, "%s", content_ptr + 1);

              current_settings->smtp_auth_state = SMTP_AUTH_STATE_UNKNOWN;
              return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
              }
            }
          else
            return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;

          break;
        case SMTP_AUTH_LOGIN:
          for (i = j + strlen_smtp_auth_types[SMTP_AUTH_LOGIN]; (i < strlen_input_line) && !isalnum((int)input_line[i]); i++);
          if ((i < strlen_input_line) &&
              isalnum((int)input_line[i]))
            {
            strncpy(current_settings->smtp_auth_response, input_line + i, MINVAL(MAX_BUF, strlen_input_line - i));
            current_settings->smtp_auth_response[MINVAL(MAX_BUF, strlen_input_line - i)] = '\0';
            current_settings->smtp_auth_state = SMTP_AUTH_STATE_RESPONSE_1_SEEN;
            }

          return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;

          break;
        case SMTP_AUTH_CRAM_MD5:
          return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
          break;
        default:
          current_settings->smtp_auth_state = SMTP_AUTH_STATE_UNKNOWN;
          return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
        }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_TLS)) &&
           (strncasecmp(SMTP_TLS, input_line, STRLEN(SMTP_TLS)) == 0))
    {
    if (current_settings->tls_state == TLS_STATE_INACTIVE)
      if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
          (current_settings->filter_grace <= FILTER_GRACE_NONE) &&
          (current_settings->rejection != NULL))
        {
        output_write_rejection(current_settings, current_settings->rejection, outbound_fd, NULL);
        current_settings->filter_grace = FILTER_GRACE_EXPIRED;
        return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
        }
      else

#ifdef HAVE_LIBSSL
        if (current_settings->tls_certificate_file != NULL)
          {
          output_write_rejection(current_settings, &tls_success, outbound_fd, NULL);
          if (tls_start(current_settings, inbound_fd, outbound_fd))
            output_writeln(current_settings, LOG_ACTION_TLS_START, -1, NULL, -1);
          else
            output_write_rejection(current_settings, &tls_failure, outbound_fd, NULL);

          return_value = FILTER_FLAG_INTERCEPT;
          }
        else
#endif /* HAVE_LIBSSL */

          return_value = FILTER_FLAG_PASS | FILTER_FLAG_TLS_CAPTURE;
    else
      {
      output_write_rejection(current_settings, &tls_failure, outbound_fd, REJECT_SEVERITY_PERMANENT);
      return_value = FILTER_FLAG_INTERCEPT;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_MAIL_FROM)) &&
           (strncasecmp(SMTP_MAIL_FROM, input_line, STRLEN(SMTP_MAIL_FROM)) == 0))
    {
    if (((strlen_target_address = find_address(input_line, strlen_input_line, target_address, MAX_BUF, &strlen_username)) > 0) &&
        (strlen_target_address < MAX_ADDRESS))
      {
      for (i = 0; i < strlen_target_address; i++)
        current_settings->sender_address[i] = tolower((int)target_address[i]);
      current_settings->sender_address[strlen_target_address] = '\0';

      if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
          (current_settings->whitelist_sender_file != NULL))
        for (i = 0; current_settings->whitelist_sender_file[i] != NULL; i++)
          if ((search_return = search_file(current_settings, current_settings->whitelist_sender_file[i], current_settings->sender_address, strlen_target_address, '@', "@.", '\0', NULL)) > 0)
            {
            current_settings->filter_action = FILTER_DECISION_DO_NOT_FILTER;
            current_settings->rejection = NULL;
            break;
            }

      if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
          (current_settings->rhswl_fqdn != NULL))
        for (i = 0; current_settings->rhswl_fqdn[i] != NULL; i++)
          if (check_rhsbl(current_settings, current_settings->sender_address + strlen_username + 1, current_settings->rhswl_fqdn[i], NULL))
            {
            current_settings->filter_action = FILTER_DECISION_DO_NOT_FILTER;
            current_settings->rejection = NULL;
            break;
            }

      if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
          (current_settings->rejection == NULL) &&
          (current_settings->blacklist_sender_file != NULL))
        for (i = 0; current_settings->blacklist_sender_file[i] != NULL; i++)
          if ((search_return = search_file(current_settings, current_settings->blacklist_sender_file[i], current_settings->sender_address, strlen_target_address, '@', "@.", '\0', NULL)) > 0)
            {
            if (current_settings->log_level >= LOG_LEVEL_DEBUG)
              {
              memcpy(&current_settings->rejection_buf, &rejection[REJECTION_SENDER_BLACKLISTED], sizeof(struct rejection_data));
              snprintf(current_settings->short_reject_message_buf, MAX_BUF, "%s(%s:%d)", rejection[REJECTION_SENDER_BLACKLISTED].short_reject_message, current_settings->blacklist_sender_file[i], search_return);
              current_settings->rejection_buf.short_reject_message = current_settings->short_reject_message_buf;
              current_settings->rejection = &current_settings->rejection_buf;
              }
            else
              current_settings->rejection = &rejection[REJECTION_SENDER_BLACKLISTED];

            break;
            }

      if ((current_settings->filter_action != FILTER_DECISION_DO_NOT_FILTER) &&
          (current_settings->rhsbl_fqdn != NULL))
        for (i = 0; current_settings->rhsbl_fqdn[i] != NULL; i++)
          if (check_rhsbl(current_settings, current_settings->sender_address + strlen_username + 1, current_settings->rhsbl_fqdn[i], &rejection[REJECTION_RHSBL]))
            {
            current_settings->filter_action = FILTER_DECISION_DO_FILTER;
            break;
            }

      if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
          (current_settings->rejection == NULL))
        {
        if (!current_settings->local_sender &&
            (current_settings->local_domains_file != NULL))
          for (i = 0; current_settings->local_domains_file[i] != NULL; i++)
            if (search_file(current_settings, current_settings->local_domains_file[i], target_address + strlen_username + 1, (strlen_target_address - strlen_username) - 1, '.', ".", '\0', NULL) > 0)
              {
              current_settings->local_sender = 1;
              break;
              }

        if (!current_settings->local_sender &&
            current_settings->check_sender_mx &&
            (((strlen_target_address - strlen_username) - 1) > 0) &&
            !dns_mx(current_settings, target_address + strlen_username + 1, NULL))
          current_settings->rejection = &rejection[REJECTION_SENDER_NO_MX];
        }
      }

    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->filter_grace <= FILTER_GRACE_AFTER_FROM) &&
        (current_settings->rejection != NULL))
      {
      output_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_NONE);
      current_settings->filter_grace = FILTER_GRACE_EXPIRED;
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_RCPT_TO)) &&
           (strncasecmp(SMTP_RCPT_TO, input_line, STRLEN(SMTP_RCPT_TO)) == 0))
    {
    if ((strlen_target_address = find_address(input_line, strlen_input_line, target_address, MAX_BUF, &strlen_username)) > 0)
      {
      for (i = 0; i < strlen_target_address; i++)
        current_settings->recipient_address[i] = tolower((int)target_address[i]);
      current_settings->recipient_address[strlen_target_address] = '\0';

      if (current_settings->local_domains_file != NULL)
        {
        local_recipient = 0;
        for (i = 0; current_settings->local_domains_file[i] != NULL; i++)
          if (search_file(current_settings, current_settings->local_domains_file[i], current_settings->recipient_address + strlen_username + 1, (strlen_target_address - strlen_username) - 1, '.', ".", '\0', NULL) > 0)
            {
            local_recipient = 1;
            break;
            }
        }
      else
        local_recipient = 1;
      }
    else
      local_recipient = 1;

    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->whitelist_recipient_file != NULL))
      for (i = 0; current_settings->whitelist_recipient_file[i] != NULL; i++)
        if ((search_return = search_file(current_settings, current_settings->whitelist_recipient_file[i], current_settings->recipient_address, strlen_target_address, '@', "@.", '@', "@")) > 0)
          {
          current_settings->filter_action = FILTER_DECISION_DO_NOT_FILTER;
          current_settings->rejection = NULL;
          break;
          }

    if (current_settings->filter_action == FILTER_DECISION_DO_FILTER)
      {
      if ((current_settings->filter_grace <= FILTER_GRACE_AFTER_TO) &&
          (current_settings->rejection != NULL))
        {
        output_write_rejection(current_settings, current_settings->rejection, outbound_fd, NULL);
        current_settings->filter_grace = FILTER_GRACE_EXPIRED;
        return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
        }
      else if (strlen_username == strlen_target_address)
        {
        current_settings->transient_rejection = &rejection[REJECTION_RCPT_TO_LOCAL];
        output_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);
        return_value = FILTER_FLAG_INTERCEPT;
        }
      else if (!local_recipient &&
               !current_settings->allow_relay)
        {
        current_settings->transient_rejection = &rejection[REJECTION_RELAYING_DENIED];
        output_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);
        return_value = FILTER_FLAG_INTERCEPT;
        }
      else if ((current_settings->max_rcpt_to > 0) &&
               (current_settings->num_rcpt_to >= current_settings->max_rcpt_to))
        {
        current_settings->transient_rejection = &rejection[REJECTION_RCPT_TO];
        output_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);
        return_value = FILTER_FLAG_INTERCEPT;
        }
      else if (current_settings->blacklist_recipient_file != NULL)
        for (i = 0; current_settings->blacklist_recipient_file[i] != NULL; i++)
          if ((search_return = search_file(current_settings, current_settings->blacklist_recipient_file[i], current_settings->recipient_address, strlen_target_address, '@', "@.", '@', "@")) > 0)
            {
            if (current_settings->log_level >= LOG_LEVEL_DEBUG)
              {
              memcpy(&current_settings->transient_rejection_buf, &rejection[REJECTION_RECIPIENT_BLACKLISTED], sizeof(struct rejection_data));
              snprintf(current_settings->transient_short_reject_message_buf, MAX_BUF, "%s(%s:%d)", rejection[REJECTION_RECIPIENT_BLACKLISTED].short_reject_message, current_settings->blacklist_recipient_file[i], search_return);
              current_settings->transient_rejection_buf.short_reject_message = current_settings->transient_short_reject_message_buf;
              current_settings->transient_rejection = &current_settings->transient_rejection_buf;
              }
            else
              current_settings->transient_rejection = &rejection[REJECTION_RECIPIENT_BLACKLISTED];

            output_write_rejection(current_settings, &rejection[REJECTION_RECIPIENT_BLACKLISTED], outbound_fd, NULL);
            return_value = FILTER_FLAG_INTERCEPT;

            break;
            }

      if ((current_settings->rejection == NULL) &&
          (current_settings->transient_rejection == NULL) &&
          (current_settings->graylist_type != GRAYLIST_NONE))
        {
        /* Figure out what kind of graylisting is wanted and if it's needed for this connection. */
        if (current_settings->graylist_type == GRAYLIST_FILTER_NONE)
          {
          continue_processing = 0;

          if (current_settings->graylist_always_ip_file != NULL)
            for (i = 0; current_settings->graylist_always_ip_file[i] != NULL; i++)
              if (search_tcprules_file(current_settings, NULL, 0, current_settings->graylist_always_ip_file[i], envp, current_settings->server_ip, current_settings->server_name, current_settings->strlen_server_name) > 0)
                {
                continue_processing = 1;
                break;
                }

          if (!continue_processing &&
              (current_settings->graylist_always_rdns_file != NULL))
            for (i = 0; current_settings->graylist_always_rdns_file[i] != NULL; i++)
              if (search_file(current_settings, current_settings->graylist_always_rdns_file[i], current_settings->server_name, current_settings->strlen_server_name, '.', ".", '\0', NULL) > 0)
                {
                continue_processing = 1;
                break;
                }

          if (!continue_processing &&
              (current_settings->graylist_always_rdns_dir != NULL))
            for (i = 0; current_settings->graylist_always_rdns_dir[i] != NULL; i++)
              if (search_domain_directory(current_settings, current_settings->graylist_always_rdns_dir[i], current_settings->server_name, current_settings->strlen_server_name) == 1)
                {
                continue_processing = 1;
                break;
                }
          }
        else
          {
          continue_processing = 1;

          if (current_settings->graylist_never_ip_file != NULL)
            for (i = 0; current_settings->graylist_never_ip_file[i] != NULL; i++)
              if (search_tcprules_file(current_settings, NULL, 0, current_settings->graylist_never_ip_file[i], envp, current_settings->server_ip, current_settings->server_name, current_settings->strlen_server_name) > 0)
                {
                continue_processing = 0;
                break;
                }

          if (continue_processing &&
              (current_settings->graylist_never_rdns_file != NULL))
            for (i = 0; current_settings->graylist_never_rdns_file[i] != NULL; i++)
              if (search_file(current_settings, current_settings->graylist_never_rdns_file[i], current_settings->server_name, current_settings->strlen_server_name, '.', ".", '\0', NULL) > 0)
                {
                continue_processing = 0;
                break;
                }

          if (continue_processing &&
              (current_settings->graylist_never_rdns_dir != NULL))
            for (i = 0; current_settings->graylist_never_rdns_dir[i] != NULL; i++)
              if (search_domain_directory(current_settings, current_settings->graylist_never_rdns_dir[i], current_settings->server_name, current_settings->strlen_server_name) == 1)
                {
                continue_processing = 0;
                break;
                }
          }

        if (continue_processing)
          {
          continue_processing = 0;
          graylist_index = -1;

          for (graylist_index = 0; current_settings->graylist_dir[graylist_index] != NULL; graylist_index++)
            if ((strlen_target_address > 0) &&
                (snprintf(current_settings->graylist_path, MAX_PATH, "%s/%s", current_settings->graylist_dir[graylist_index], canonicalize_path(tmp_recipient_domain, MAX_PATH, target_address + strlen_username + 1, (strlen_target_address - strlen_username) - 1)) < MAX_PATH) &&
                (stat(current_settings->graylist_path, &tmp_stat) == 0) &&
                (S_ISDIR(tmp_stat.st_mode)))
              {
              continue_processing = 1;
              break;
              }
          }

        /* Construct the path to the recipient's domain's graylist folder and check if it exists */
        if (continue_processing)
          {
          if ((snprintf(current_settings->graylist_path, MAX_PATH, "%s/%s/%s", current_settings->graylist_dir[graylist_index], tmp_recipient_domain, canonicalize_path(tmp_recipient_username, MAX_PATH, target_address, strlen_username)) < MAX_PATH) &&
              (stat(current_settings->graylist_path, &tmp_stat) != 0))
            {
            if (errno == ENOENT)
              {
              if (mkdir(current_settings->graylist_path, MKDIR_MODE) != 0)
                {
                continue_processing = 0;
                SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_MKDIR "%s: %s", current_settings->graylist_path, strerror(errno));
                }
              }
            else
              {
              continue_processing = 0;
              SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_STAT "%s: %s", current_settings->graylist_path, strerror(errno));
              }
            }

          /* Construct the path to the sender->user file and check if it exists */
          if (continue_processing &&
              (snprintf(current_settings->graylist_path, MAX_PATH, "%s/%s/%s/%s", current_settings->graylist_dir[graylist_index], tmp_recipient_domain, tmp_recipient_username, (current_settings->sender_address[0] != '\0') ? canonicalize_path(tmp_sender_address, MAX_PATH, current_settings->sender_address, strlen(current_settings->sender_address)) : SENDER_ADDRESS_NONE) < MAX_PATH) &&
              (stat(current_settings->graylist_path, &tmp_stat) == 0))
            /* Check that the file is a "regular" file and the age matches the command line options */
            if ((S_ISREG(tmp_stat.st_mode)) &&
                (((current_settings->graylist_min_secs > 0) &&
                  (((tmp_stat.st_mtime + current_settings->graylist_min_secs) <= time(NULL)) ||
                   (tmp_stat.st_size > 0))) ||
                 (current_settings->graylist_min_secs == 0)) &&
                (((current_settings->graylist_max_secs > 0) &&
                  ((tmp_stat.st_mtime + current_settings->graylist_max_secs) >= time(NULL))) ||
                 (current_settings->graylist_max_secs == 0)))
              {
              /* User was graylisted but is now allowed.  Log the name of their sending host to the file. */
              if (current_settings->strlen_server_name > 0)
                snprintf(log_entry, MAX_BUF, "%s %s", current_settings->server_ip, current_settings->server_name);
              else
                snprintf(log_entry, MAX_BUF, "%s", current_settings->server_ip);

              search_results = search_file(current_settings, current_settings->graylist_path, log_entry, strlen(log_entry), '\0', NULL, '\0', NULL);
              if ((tmp_file = fopen(current_settings->graylist_path, "a")) != NULL)
                {
                chmod(current_settings->graylist_path, CHMOD_MODE);

                if (search_results <= 0)
                  fprintf(tmp_file, "%s\n", log_entry);

                fclose(tmp_file);
                }
              else
                SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_GRAYLIST_FILE "%s: %s", current_settings->graylist_path, strerror(errno));
              }
            else
              {
              /* The file's age is outside the bounds of the command line options. */
              return_value = FILTER_FLAG_INTERCEPT;

              /* The file was too old -- truncate it. */
              if ((current_settings->graylist_max_secs > 0) &&
                  ((tmp_stat.st_mtime + current_settings->graylist_max_secs) <= time(NULL)))
                {
                if ((tmp_file = fopen(current_settings->graylist_path, "w")) != NULL)
                  {
                  chmod(current_settings->graylist_path, CHMOD_MODE);

                  fclose(tmp_file);
                  }
                else
                  {
                  return_value = FILTER_FLAG_PASS;
                  SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_GRAYLIST_FILE "%s: %s", current_settings->graylist_path, strerror(errno));
                  }
                }

              if ((return_value & FILTER_MASK_PASS) != FILTER_FLAG_PASS)
                {
                current_settings->transient_rejection = &rejection[REJECTION_GRAYLISTED];
                output_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);
                }
              }
          /* Sender->user graylist file does not exist -- create it. */
          else if (continue_processing &&
                   ((tmp_file = fopen(current_settings->graylist_path, "w")) != NULL))
            {
            chmod(current_settings->graylist_path, CHMOD_MODE);

            fclose(tmp_file);

            current_settings->transient_rejection = &rejection[REJECTION_GRAYLISTED];
            output_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);

            return_value = FILTER_FLAG_INTERCEPT;
            }
          else if (continue_processing)
            SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_GRAYLIST_FILE "%s: %s", current_settings->graylist_path, strerror(errno));
          }
        }
      }

    if ((return_value & FILTER_MASK_PASS) == FILTER_FLAG_PASS)
      return_value = FILTER_FLAG_PASS | FILTER_FLAG_RCPT_CAPTURE;
    else
      {
      SPAMDYKE_LOG_INFO(current_settings, "%s from: %s to: %s origin_ip: %s origin_rdns: %s auth: %s", (current_settings->transient_rejection != NULL) ? current_settings->transient_rejection->short_reject_message : ((current_settings->rejection != NULL) ? current_settings->rejection->short_reject_message : SHORT_SUCCESS), (current_settings->sender_address[0] != '\0') ? current_settings->sender_address : LOG_MISSING_DATA, (current_settings->recipient_address[0] != '\0') ? current_settings->recipient_address : LOG_MISSING_DATA, (current_settings->server_ip != NULL) ? current_settings->server_ip : LOG_MISSING_DATA, (current_settings->strlen_server_name > 0) ? current_settings->server_name : LOG_MISSING_DATA, (current_settings->smtp_auth_username[0] != '\0') ? current_settings->smtp_auth_username : LOG_MISSING_DATA);
      current_settings->transient_rejection = NULL;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_DATA)) &&
           (strncasecmp(SMTP_DATA, input_line, STRLEN(SMTP_DATA)) == 0))
    {
    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->filter_grace <= FILTER_GRACE_AFTER_TO) &&
        (current_settings->rejection != NULL))
      {
      output_write_rejection(current_settings, current_settings->rejection, outbound_fd, NULL);
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    else if (current_settings->num_rcpt_to == 0)
      {
      output_write_rejection(current_settings, &rejection[REJECTION_ZERO_RECIPIENTS], STDOUT_FD, NULL);
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_CONTINUE;
      }
    else
      current_settings->inside_data = 1;
    }
  else if ((strlen_input_line >= STRLEN(SMTP_QUIT)) &&
           (strncasecmp(SMTP_QUIT, input_line, STRLEN(SMTP_QUIT)) == 0))
    {
    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->filter_grace == FILTER_GRACE_EXPIRED) &&
        (current_settings->rejection != NULL))
      {
      output_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_QUIT);
      return_value = FILTER_FLAG_QUIT;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_RSET)) &&
           (strncasecmp(SMTP_RSET, input_line, STRLEN(SMTP_RSET)) == 0))
    {
    if ((current_settings->filter_action != FILTER_DECISION_DO_FILTER) ||
        (current_settings->filter_grace != FILTER_GRACE_EXPIRED) ||
        (current_settings->rejection == NULL))
      {
      current_settings->num_rcpt_to = 0;
      current_settings->local_sender = 1;
      current_settings->sender_address[0] = '\0';
      current_settings->recipient_address[0] = '\0';
      current_settings->graylist_path[0] = '\0';
      }
    else
      {
      output_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_PERMANENT);
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    }
  else if ((strlen_input_line < STRLEN(SMTP_QUIT)) ||
           (strncasecmp(SMTP_QUIT, input_line, STRLEN(SMTP_QUIT)) != 0))
    {
    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->filter_grace == FILTER_GRACE_EXPIRED) &&
        (current_settings->rejection != NULL))
      {
      output_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_PERMANENT);
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    //FIXME: Not sure I like this here.  What if input for an unknown auth protocol looks like an SMTP command?
    else if (current_settings->smtp_auth_state == SMTP_AUTH_STATE_UNKNOWN)
      return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
    }

  return(return_value);
  }

/*
 * Expects:
 *   inbound_fd == pointer to input file descriptor from child process
 *   outbound_fd == pointer to output file descriptor to child process
 */
void middleman(int *inbound_fd, int *outbound_fd, char line_terminator, struct filter_settings *current_settings, int child_pid, char *envp[])
  {
  char buf_input[MAX_BUF + 1];
  char buf_childinput[MAX_BUF + 1];
  int strlen_buf_input;
  int strlen_buf_childinput;
  int strlen_buf_trim;
  int tmp_strlen;
  fd_set listen_set;
  struct timeval listen_timeout;
  int more_stdinput;
  int more_childinput;
  int discard_childinput;
  long connection_start;
  long command_start;
  long current_time;
  int filter_return;
  int network_ready;
  int added_ehlo_continuation;
  int read_result;
  char *start_buf_input;
  char *start_buf_childinput;
  char *next_terminator;
  int child_response_needed;
  int max_fd;
  int select_result;
  int usable_buf_input;
  int timeout_printed;

  strlen_buf_input = 0;
  usable_buf_input = 0;
  buf_input[0] = '\0';
  start_buf_input = NULL;
  buf_childinput[0] = '\0';
  start_buf_childinput = NULL;
  strlen_buf_childinput = 0;
  filter_return = FILTER_FLAG_PASS;
  child_response_needed = 0;
  timeout_printed = 0;

  more_stdinput = 1;
  more_childinput = ((*inbound_fd >= 0) && (*outbound_fd >= 0)) ? 1 : 0;
  discard_childinput = !more_childinput;

  connection_start = time(NULL);
  command_start = time(NULL);

  /*
   * Deciding when to stop looping and disconnect is tricky and this code gets
   * updated with almost every version.  more_childinput should indicate if
   * the qmail child process is still running and thus if spamdyke should
   * continue to accept input from it.  discard_childinput should indicate if
   * spamdyke is discarding qmail's output.  This happens when a filter is
   * triggered and qmail is cut off.  more_stdinput should indicate if the
   * remote host is still connected and sending data.
   */
  while (more_childinput ||
         (discard_childinput &&
          more_stdinput))
    {
    current_time = time(NULL);

    if (!timeout_printed &&
        (((current_settings->timeout_connection > 0) &&
          ((current_time - connection_start) > current_settings->timeout_connection)) ||
         ((current_settings->timeout_command > 0) &&
          ((current_time - command_start) > current_settings->timeout_command))))
      {
      if (*outbound_fd >= 0)
        {
        output_writeln(current_settings, LOG_ACTION_FILTER_TO, *outbound_fd, CHILD_QUIT, STRLEN(CHILD_QUIT));
        close(*outbound_fd);
        *outbound_fd = -1;
        }

      output_write_rejection(current_settings, &rejection[REJECTION_TIMEOUT], STDOUT_FD, NULL);

      more_stdinput = 0;
      discard_childinput = 1;
      child_response_needed = 0;

      SPAMDYKE_LOG_INFO(current_settings, "%s from: %s to: %s origin_ip: %s origin_rdns: %s auth: %s reason: %s", rejection[REJECTION_TIMEOUT].short_reject_message, (current_settings->sender_address[0] != '\0') ? current_settings->sender_address : LOG_MISSING_DATA, (current_settings->recipient_address[0] != '\0') ? current_settings->recipient_address : LOG_MISSING_DATA, current_settings->server_ip, (current_settings->strlen_server_name > 0) ? current_settings->server_name : LOG_MISSING_DATA, (current_settings->smtp_auth_username[0] != '\0') ? current_settings->smtp_auth_username : LOG_MISSING_DATA, (current_settings->rejection != NULL) ? current_settings->rejection->short_reject_message : LOG_MISSING_DATA);

      timeout_printed = 1;
      }

    listen_timeout.tv_sec = 0;
    listen_timeout.tv_usec = 0;

    if ((current_settings->timeout_connection > 0) ||
        (current_settings->timeout_command > 0))
      {
      if (current_settings->timeout_command > 0)
        listen_timeout.tv_sec = MAXVAL(0, current_settings->timeout_command - (current_time - command_start));

      if (current_settings->timeout_connection > 0)
        listen_timeout.tv_sec = MAXVAL(0, MINVAL(listen_timeout.tv_sec, current_settings->timeout_connection - (current_time - connection_start)));

      if (listen_timeout.tv_sec > MAX_SELECT_SECS_TIMEOUT)
        {
        listen_timeout.tv_sec = MAX_SELECT_SECS_TIMEOUT;
        listen_timeout.tv_usec = MAX_SELECT_USECS_TIMEOUT;
        }
      else if ((listen_timeout.tv_sec == 0) ||
               (listen_timeout.tv_sec < MIN_SELECT_SECS_TIMEOUT))
        {
        listen_timeout.tv_sec = MIN_SELECT_SECS_TIMEOUT;
        listen_timeout.tv_usec = MIN_SELECT_USECS_TIMEOUT;
        }
      }
    else
      {
      listen_timeout.tv_sec = SELECT_SECS_NO_TIMEOUT;
      listen_timeout.tv_usec = SELECT_USECS_NO_TIMEOUT;
      }

    FD_ZERO(&listen_set);
    max_fd = 0;
    if (more_stdinput &&
        !child_response_needed)
      {
      FD_SET(STDIN_FD, &listen_set);
      max_fd = STDIN_FD;
      }
    if (more_childinput)
      {
      FD_SET(*inbound_fd, &listen_set);
      max_fd = *inbound_fd;
      }

    /*
     * The NETWORK_CAN_READ macro only has meaning when TLS is available;
     * sometimes OpenSSL collects traffic in its buffer and requires multiple
     * reads to deliver it all.  select()ing on the socket in that state will
     * block because the data is in OpenSSL's buffer, not the network layer.
     */
    network_ready = 0;
    select_result = -1;
    if ((usable_buf_input &&
         !child_response_needed) ||
        (strlen_buf_childinput > 0) ||
        (!child_response_needed &&
         ((network_ready = NETWORK_CAN_READ(current_settings)) > 0)) ||
        ((select_result = select(max_fd + 1, &listen_set, NULL, NULL, &listen_timeout)) > 0))
      {
      if (select_result == -1)
        {
        if (network_ready == 0)
          {
          listen_timeout.tv_sec = 0;
          listen_timeout.tv_usec = 0;
          select(max_fd + 1, &listen_set, NULL, NULL, &listen_timeout);
          }
        else
          FD_ZERO(&listen_set);
        }

      next_terminator = NULL;

      /*
       * Handle input from the child process
       */
      if ((*inbound_fd >= 0) &&
          FD_ISSET(*inbound_fd, &listen_set))
        {
        if ((read_result = read(*inbound_fd, buf_childinput + strlen_buf_childinput, MAX_BUF - strlen_buf_childinput)) > 0)
          {
          strlen_buf_childinput += read_result;
          buf_childinput[strlen_buf_childinput] = '\0';

          if (start_buf_childinput == NULL)
            start_buf_childinput = buf_childinput;
          }
        else
          {
          more_childinput = 0;
          close(*inbound_fd);
          *inbound_fd = -1;
          }
        }

      if (strlen_buf_childinput > 0)
        {
        while ((line_terminator != '\0') &&
               (current_settings->tls_state != TLS_STATE_ACTIVE_PASSTHROUGH) &&
               ((next_terminator = strchr(start_buf_childinput, line_terminator)) != NULL))
          {
          next_terminator++;
          child_response_needed = 0;
          command_start = time(NULL);

          strlen_buf_trim = next_terminator - start_buf_childinput;
          while ((strlen_buf_trim > 0) &&
                 ((start_buf_childinput[strlen_buf_trim - 1] == '\r') ||
                  (start_buf_childinput[strlen_buf_trim - 1] == '\n')))
            strlen_buf_trim--;

          if (!discard_childinput)
            {
            if ((line_terminator != '\0') &&
                (((filter_return & FILTER_MASK_AUTH) != FILTER_FLAG_AUTH_NONE) ||
                 ((filter_return & FILTER_MASK_TLS) != FILTER_FLAG_TLS_NONE) ||
                 ((filter_return & FILTER_MASK_RCPT) != FILTER_FLAG_RCPT_NONE)) &&
                (strlen_buf_trim > 0))
              {
              if ((filter_return & FILTER_MASK_RCPT) == FILTER_FLAG_RCPT_CAPTURE)
                {
                /* The client said "RCPT TO" and spamdyke didn't block it.  Capture qmail's response. */
                if ((strlen_buf_trim >= STRLEN(SMTP_RCPT_SUCCESS)) &&
                    (strncasecmp(start_buf_childinput, SMTP_RCPT_SUCCESS, STRLEN(SMTP_RCPT_SUCCESS)) == 0))
                  {
                  /* qmail said "250".  Log success. */
                  current_settings->num_rcpt_to++;
                  }
                else
                  {
                  /* qmail did not say "250".  Log a rejection. */
                  current_settings->transient_rejection = &rejection[REJECTION_OTHER];
                  }

                output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                filter_return ^= FILTER_FLAG_RCPT_CAPTURE;

                SPAMDYKE_LOG_INFO(current_settings, "%s from: %s to: %s origin_ip: %s origin_rdns: %s auth: %s", (current_settings->transient_rejection != NULL) ? current_settings->transient_rejection->short_reject_message : ((current_settings->rejection != NULL) ? current_settings->rejection->short_reject_message : SHORT_SUCCESS), (current_settings->sender_address[0] != '\0') ? current_settings->sender_address : LOG_MISSING_DATA, (current_settings->recipient_address[0] != '\0') ? current_settings->recipient_address : LOG_MISSING_DATA, (current_settings->server_ip != NULL) ? current_settings->server_ip : LOG_MISSING_DATA, (current_settings->strlen_server_name > 0) ? current_settings->server_name : LOG_MISSING_DATA, (current_settings->smtp_auth_username[0] != '\0') ? current_settings->smtp_auth_username : LOG_MISSING_DATA);
                current_settings->transient_rejection = NULL;
                }
              else if ((((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD) ||
                        ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD) ||
                        ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_REMOVE)) &&
                       (strlen_buf_trim >= STRLEN(SMTP_EHLO_SUCCESS)) &&
                       (strncasecmp(start_buf_childinput, SMTP_EHLO_SUCCESS, STRLEN(SMTP_EHLO_SUCCESS)) == 0))
                {
                /* The server said "250". */
                added_ehlo_continuation = 0;

                if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD) &&
                    (strlen_buf_trim >= (STRLEN(SMTP_EHLO_SUCCESS) + STRLEN(SMTP_EHLO_AUTH) + 1)) &&
                    (strncasecmp(start_buf_childinput + STRLEN(SMTP_EHLO_SUCCESS) + 1, SMTP_EHLO_AUTH, STRLEN(SMTP_EHLO_AUTH)) == 0))
                  {
                  /* The server said "250 AUTH ".  Hopefully it doesn't support an algorithm spamdyke can't handle... */
                  if (((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD) &&
                      (start_buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] != SMTP_CONTINUATION))
                    {
                    start_buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] = SMTP_CONTINUATION;
                    added_ehlo_continuation = 1;
                    }

                  strlen_buf_trim = 0;
                  current_settings->smtp_auth_origin = SMTP_AUTH_ORIGIN_CHILD;
                  filter_return ^= FILTER_FLAG_AUTH_ADD;
                  output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                  }
                else if (((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD) &&
                        (strlen_buf_trim >= (STRLEN(SMTP_EHLO_SUCCESS) + STRLEN(SMTP_EHLO_TLS) + 1)) &&
                        (strncasecmp(start_buf_childinput + STRLEN(SMTP_EHLO_SUCCESS) + 1, SMTP_EHLO_TLS, STRLEN(SMTP_EHLO_TLS)) == 0))
                  {
                  /* The server said "250 STARTTLS". */
                  if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD) &&
                      (start_buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] != SMTP_CONTINUATION))
                    {
                    start_buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] = SMTP_CONTINUATION;
                    added_ehlo_continuation = 1;
                    }

                  strlen_buf_trim = 0;
                  filter_return ^= FILTER_FLAG_TLS_ADD;
                  output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                  }
                else if (((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_REMOVE) &&
                        (strlen_buf_trim >= (STRLEN(SMTP_EHLO_SUCCESS) + STRLEN(SMTP_EHLO_TLS) + 1)) &&
                        (strncasecmp(start_buf_childinput + STRLEN(SMTP_EHLO_SUCCESS) + 1, SMTP_EHLO_TLS, STRLEN(SMTP_EHLO_TLS)) == 0))
                  {
                  /* The server said "250 STARTTLS".  Remove it. */
                  output_writeln(current_settings, LOG_ACTION_CHILD_FROM_DISCARDED, -1, start_buf_childinput, next_terminator - start_buf_childinput);

                  if (((filter_return & FILTER_MASK_AUTH) != FILTER_FLAG_AUTH_ADD) &&
                      (start_buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] != SMTP_CONTINUATION))
                    {
                    /* The server said "250 STARTTLS".  Remove it and replace it with something bogus because the previous line was a continuation. */
                    output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_SUCCESS, STRLEN(SMTP_EHLO_SUCCESS));
                    output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_STR_DONE, STRLEN(SMTP_STR_DONE));
                    output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_NOTHING_INSERT, STRLEN(SMTP_EHLO_NOTHING_INSERT));
                    }

                  strlen_buf_trim = 0;
                  filter_return ^= FILTER_FLAG_TLS_REMOVE;
                  }

                if (strlen_buf_trim > 0)
                  {
                  /* This line is not "AUTH" or "STARTTLS". */
                  if ((strlen_buf_trim >= (STRLEN(SMTP_EHLO_SUCCESS) + 1)) &&
                      (start_buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] != SMTP_CONTINUATION) &&
                      (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD) ||
                       ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD)))
                    {
                    /* The server said "250 " without saying "AUTH" and/or "STARTTLS", so change the last line to "250-". */
                    start_buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] = SMTP_CONTINUATION;
                    added_ehlo_continuation = 1;
                    }

                  output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                  strlen_buf_trim = 0;
                  }

                if ((((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD) ||
                     ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD) ||
                     ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_REMOVE)) &&
                     added_ehlo_continuation)
                  {
                  if ((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD)
                    {
                    filter_return ^= FILTER_FLAG_AUTH_ADD;
                    current_settings->smtp_auth_origin = SMTP_AUTH_ORIGIN_SPAMDYKE;

                    /* Add a "250 AUTH" (or "250-AUTH") line. */
                    output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_SUCCESS, STRLEN(SMTP_EHLO_SUCCESS));

                    if ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD)
                      output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_STR_CONTINUATION, STRLEN(SMTP_STR_CONTINUATION));
                    else
                      output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_STR_DONE, STRLEN(SMTP_STR_DONE));

                    if (current_settings->smtp_auth_encryption)
                      output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_AUTH_INSERT_ENCRYPTION, STRLEN(SMTP_EHLO_AUTH_INSERT_ENCRYPTION));
                    else
                      output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_AUTH_INSERT_CLEAR, STRLEN(SMTP_EHLO_AUTH_INSERT_CLEAR));
                    }

                  if ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD)
                    {
                    filter_return ^= FILTER_FLAG_TLS_ADD;

                    /* Add a "250 STARTTLS" line. */
                    output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_SUCCESS, STRLEN(SMTP_EHLO_SUCCESS));
                    output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_STR_DONE, STRLEN(SMTP_STR_DONE));
                    output_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_TLS_INSERT, STRLEN(SMTP_EHLO_TLS_INSERT));
                    }
                  else if ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_REMOVE)
                    filter_return ^= FILTER_FLAG_TLS_REMOVE;
                  }
                else if (strlen_buf_trim > 0)
                  output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                }
              else if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE) &&
                       (current_settings->smtp_auth_state == SMTP_AUTH_STATE_CMD_SEEN) &&
                       (strlen_buf_trim >= STRLEN(SMTP_AUTH_CHALLENGE)) &&
                       (strncasecmp(start_buf_childinput, SMTP_AUTH_CHALLENGE, STRLEN(SMTP_AUTH_CHALLENGE)) == 0))
                {
                strncpy(current_settings->smtp_auth_challenge, start_buf_childinput, MINVAL(MAX_BUF, strlen_buf_trim));
                current_settings->smtp_auth_challenge[MINVAL(MAX_BUF, strlen_buf_trim)] = '\0';
                current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_1_SENT;
                output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                }
              else if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE) &&
                       (current_settings->smtp_auth_state == SMTP_AUTH_STATE_RESPONSE_1_SEEN) &&
                       (strlen_buf_trim >= STRLEN(SMTP_AUTH_CHALLENGE)) &&
                       (strncasecmp(start_buf_childinput, SMTP_AUTH_CHALLENGE, STRLEN(SMTP_AUTH_CHALLENGE)) == 0))
                {
                current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_2_SENT;
                output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                }
              else if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE) &&
                       (current_settings->smtp_auth_state == SMTP_AUTH_STATE_RESPONSE_2_SEEN) &&
                       (strlen_buf_trim >= STRLEN(SMTP_AUTH_SUCCESS)) &&
                       (strncasecmp(start_buf_childinput, SMTP_AUTH_SUCCESS, STRLEN(SMTP_AUTH_SUCCESS)) == 0))
                {
                current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
                current_settings->filter_action = FILTER_DECISION_DO_NOT_FILTER;
                current_settings->rejection = NULL;
                filter_return ^= FILTER_FLAG_AUTH_CAPTURE;
                output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                }
              else if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE) &&
                       (current_settings->smtp_auth_state == SMTP_AUTH_STATE_UNKNOWN) &&
                       (strlen_buf_trim >= STRLEN(SMTP_AUTH_PROMPT)) &&
                       (strncasecmp(start_buf_childinput, SMTP_AUTH_PROMPT, STRLEN(SMTP_AUTH_PROMPT)) == 0))
                {
                // Continue watching the AUTH conversation.
                output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                }
              else if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE) &&
                       (current_settings->smtp_auth_state == SMTP_AUTH_STATE_UNKNOWN) &&
                       (strlen_buf_trim >= STRLEN(SMTP_AUTH_SUCCESS)) &&
                       (strncasecmp(start_buf_childinput, SMTP_AUTH_SUCCESS, STRLEN(SMTP_AUTH_SUCCESS)) == 0))
                {
                tmp_strlen = strlen(current_settings->smtp_auth_response);
                strncpy(current_settings->smtp_auth_username, current_settings->smtp_auth_response, MINVAL(tmp_strlen, MAX_BUF));
                current_settings->smtp_auth_username[MINVAL(tmp_strlen, MAX_BUF)] = '\0';

                current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
                current_settings->filter_action = FILTER_DECISION_DO_NOT_FILTER;
                current_settings->rejection = NULL;
                filter_return ^= FILTER_FLAG_AUTH_CAPTURE;
                output_writeln(current_settings, LOG_ACTION_AUTH_SUCCESS, -1, NULL, -1);
                output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                }
              else if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE) &&
                       (current_settings->smtp_auth_state == SMTP_AUTH_STATE_UNKNOWN) &&
                       (strlen_buf_trim >= STRLEN(SMTP_AUTH_FAILURE)) &&
                       (strncasecmp(start_buf_childinput, SMTP_AUTH_FAILURE, STRLEN(SMTP_AUTH_FAILURE)) == 0))
                {
                current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
                current_settings->filter_action = FILTER_DECISION_DO_FILTER;
                filter_return ^= FILTER_FLAG_AUTH_CAPTURE;
                output_writeln(current_settings, LOG_ACTION_AUTH_FAILURE, -1, (current_settings->smtp_auth_response[0] != '\0') ? current_settings->smtp_auth_response : LOG_MISSING_DATA, -1);
                output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                }
              else if (((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_CAPTURE) &&
                       (strlen_buf_trim >= STRLEN(SMTP_TLS_SUCCESS)) &&
                       (strncasecmp(start_buf_childinput, SMTP_TLS_SUCCESS, STRLEN(SMTP_TLS_SUCCESS)) == 0))
                {
                output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                current_settings->tls_state = TLS_STATE_ACTIVE_PASSTHROUGH;
                filter_return ^= FILTER_FLAG_TLS_CAPTURE;
                output_writeln(current_settings, LOG_ACTION_TLS_PASSTHROUGH_START, -1, NULL, -1);
                output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, next_terminator, strlen_buf_childinput - (next_terminator - start_buf_childinput));
                next_terminator = NULL;
                start_buf_childinput = NULL;
                strlen_buf_childinput = 0;
                child_response_needed = 0;
                line_terminator = '\0';
                buf_childinput[strlen_buf_childinput] = '\0';
                }
              else
                {
                if ((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE)
                  {
                  filter_return ^= FILTER_FLAG_AUTH_CAPTURE;
                  current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
                  }
                else if ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_CAPTURE)
                  {
                  filter_return ^= FILTER_FLAG_TLS_CAPTURE;
                  current_settings->filter_action = FILTER_DECISION_DO_FILTER;
                  }

                output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
                }
              }
            else
              output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, next_terminator - start_buf_childinput);
            }
          else
            output_writeln(current_settings, LOG_ACTION_CHILD_FROM_DISCARDED, -1, start_buf_childinput, next_terminator - start_buf_childinput);

          start_buf_childinput = next_terminator;
          }

        if ((line_terminator != '\0') &&
            (current_settings->tls_state != TLS_STATE_ACTIVE_PASSTHROUGH))
          {
          if ((start_buf_childinput - buf_childinput) >= strlen_buf_childinput)
            {
            /* Used entire buffer, reset pointers */
            start_buf_childinput = NULL;
            strlen_buf_childinput = 0;
            }
          else if (strlen_buf_childinput >= MAX_BUF)
            {
            /* Buffer is full */
            if (start_buf_childinput > buf_childinput)
              {
              /* Unused space at the start, move data from the end to the start */
              strlen_buf_childinput -= start_buf_childinput - buf_childinput;
              memmove(buf_childinput, start_buf_childinput, strlen_buf_childinput);
              start_buf_childinput = buf_childinput;
              }
            else
              {
              /* Buffer is completely full, output data */
              output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, start_buf_childinput, strlen_buf_childinput);
              start_buf_childinput = NULL;
              strlen_buf_childinput = 0;
              }
            }
          else if ((next_terminator == NULL) &&
                   (start_buf_childinput != buf_childinput))
            {
            /* Buffer is not full and no terminator found, move data */
            strlen_buf_childinput -= start_buf_childinput - buf_childinput;
            memmove(buf_childinput, start_buf_childinput, strlen_buf_childinput);
            start_buf_childinput = NULL;
            }
          }
        else if (strlen_buf_childinput > 0)
          {
          output_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
          start_buf_childinput = NULL;
          strlen_buf_childinput = 0;
          }

        buf_childinput[strlen_buf_childinput] = '\0';
        }

      next_terminator = NULL;

      /*
       * Handle input from the remote server
       */
      if (FD_ISSET(STDIN_FD, &listen_set) ||
          NETWORK_CAN_READ(current_settings))
        {
        if ((read_result = NETWORK_READ(current_settings, STDIN_FD, buf_input + strlen_buf_input, MAX_BUF - strlen_buf_input)) > 0)
          {
          strlen_buf_input += read_result;
          buf_input[strlen_buf_input] = '\0';
          usable_buf_input = 1;

          if (current_settings->tls_state != TLS_STATE_ACTIVE_PASSTHROUGH)
            {
            while ((strlen_buf_input > 0) &&
                   (buf_input[strlen_buf_input - 1] == '\0'))
              strlen_buf_input--;

            if ((strlen_buf_input > 0) &&
                (start_buf_input == NULL))
              start_buf_input = buf_input;
            }

          command_start = time(NULL);
          }
        else if (!NETWORK_CAN_READ(current_settings))
          {
          if (*outbound_fd >= 0)
            {
            if ((start_buf_input == NULL) &&
                (strlen_buf_input > 0))
              start_buf_input = buf_input;

            if ((start_buf_input != NULL) &&
                ((strlen_buf_input - (start_buf_input - buf_input)) > 0))
              {
              output_writeln(current_settings, LOG_ACTION_REMOTE_FROM, *outbound_fd, start_buf_input, strlen_buf_input - (start_buf_input - buf_input));
              strlen_buf_input = 0;
              buf_input[strlen_buf_input] = '\0';
              start_buf_input = NULL;
              }

            close(*outbound_fd);
            *outbound_fd = -1;
            }

          more_stdinput = 0;
          child_response_needed = 0;
          }
        }

      if ((strlen_buf_input > 0) &&
          !child_response_needed)
        {
        if ((line_terminator != '\0') &&
            (current_settings->tls_state != TLS_STATE_ACTIVE_PASSTHROUGH))
          {
          if (start_buf_input == NULL)
            start_buf_input = buf_input;

          if (((next_terminator = strchr(start_buf_input, line_terminator)) != NULL) ||
              (strlen_buf_input == MAX_BUF))
            {
            if (next_terminator != NULL)
              {
              next_terminator++;

              strlen_buf_trim = next_terminator - start_buf_input;
              while ((strlen_buf_trim > 0) &&
                     ((start_buf_input[strlen_buf_trim - 1] == '\r') ||
                      (start_buf_input[strlen_buf_trim - 1] == '\n')))
                strlen_buf_trim--;
              }
            else
              {
              next_terminator = buf_input + strlen_buf_input;
              strlen_buf_trim = strlen_buf_input;
              }

            output_writeln(current_settings, LOG_ACTION_REMOTE_FROM, -1, start_buf_input, next_terminator - start_buf_input);

            filter_return = smtp_filter(STDIN_FD, STDOUT_FD, start_buf_input, strlen_buf_trim, current_settings, envp);
            switch (filter_return & FILTER_MASK_PASS)
              {
              case FILTER_FLAG_PASS:
                child_response_needed = (((filter_return & FILTER_MASK_CHILD_RESPONSE) == FILTER_FLAG_CHILD_RESPONSE_NEEDED) && !discard_childinput) ? 1 : 0;
                output_writeln(current_settings, LOG_ACTION_NONE, *outbound_fd, start_buf_input, next_terminator - start_buf_input);
                break;
              case FILTER_FLAG_QUIT:
                more_stdinput = 0;
                discard_childinput = 1;
                child_response_needed = 0;

                break;
              case FILTER_FLAG_INTERCEPT:
                if (((filter_return & FILTER_MASK_CHILD) == FILTER_FLAG_CHILD_QUIT) &&
                    (current_settings->rejection != NULL) &&
                    (*outbound_fd >= 0) &&
                    !discard_childinput)
                  {
                  output_writeln(current_settings, LOG_ACTION_FILTER_TO, *outbound_fd, CHILD_QUIT, STRLEN(CHILD_QUIT));
                  close(*outbound_fd);
                  *outbound_fd = -1;

                  discard_childinput = 1;
                  }
              }

            start_buf_input = next_terminator;
            }
          else
            usable_buf_input = 0;

          if ((start_buf_input - buf_input) >= strlen_buf_input)
            {
            /* Used entire buffer, reset pointers */
            start_buf_input = NULL;
            strlen_buf_input = 0;
            }
          else if (strlen_buf_input >= MAX_BUF)
            {
            /* Buffer is full */
            if (start_buf_input > buf_input)
              {
              /* Unused space at the start, move data from the end to the start */
              strlen_buf_input -= start_buf_input - buf_input;
              memmove(buf_input, start_buf_input, strlen_buf_input);
              start_buf_input = buf_input;
              }
            else
              {
              /* Buffer is completely full, output data */
              output_writeln(current_settings, LOG_ACTION_REMOTE_FROM, *outbound_fd, start_buf_input, strlen_buf_input);
              start_buf_input = NULL;
              strlen_buf_input = 0;
              }
            }
          else if ((next_terminator == NULL) &&
                   (start_buf_input != buf_input))
            {
            /* Buffer is not full and no terminator found, move data */
            strlen_buf_input -= start_buf_input - buf_input;
            memmove(buf_input, start_buf_input, strlen_buf_input);
            start_buf_input = NULL;
            }
          }
        else
          {
          output_writeln(current_settings, LOG_ACTION_REMOTE_FROM, *outbound_fd, buf_input, strlen_buf_input);
          start_buf_input = NULL;
          strlen_buf_input = 0;
          }

        buf_input[strlen_buf_input] = '\0';
        }

      if (strlen_buf_input == 0)
        usable_buf_input = 0;
      }

    if (child_pid != -1)
      waitpid(child_pid, NULL, WNOHANG);
    }

  return;
  }

/*
 * Return value:
 *   ERROR: 0
 *   NOT FOUND: 0
 *   SUCCESS: 1
 */
int search_ip(char *target_ip, int strlen_target_ip, char *target_name, int strlen_target_name)
  {
  int j;
  int k;

  j = 0;
  k = 0;
  while (k < strlen_target_name)
    if (j < strlen_target_ip)
      if ((target_ip[j] >= '0') &&
          (target_ip[j] <= '9'))
        {
        if (target_ip[j] == target_name[k])
          j++;
        else
          j = 0;

        k++;
        }
      else
        {
        while ((k < strlen_target_name) &&
               ((target_name[k] < '0') ||
                (target_name[k] > '9')))
          k++;

        j++;
        }
    else
      break;

  return(((j > 0) && (j == strlen_target_ip)) ? 1 : 0);
  }

/*
 * Return value:
 *   ERROR: -1
 *   NOT FOUND: 0
 *   SUCCESS: 1
 */
int is_ip_in_name(char *target_ip, char *target_name)
  {
  int return_value;
  char ip_octets[4][4];
  int ip_ints[4];
  char tmp_ip[MAX_IP + 1];
  int tmp_strlen;
  struct in_addr tmp_addr;
  int strlen_target_ip;
  int strlen_target_name;

  return_value = -1;

  if ((target_ip != NULL) &&
      ((strlen_target_ip = strlen(target_ip)) > 0) &&
      (sscanf(target_ip, "%3[0-9].%3[0-9].%3[0-9].%3[0-9]", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]) == 4) &&
      (sscanf(ip_octets[0], "%d", &ip_ints[0]) == 1) &&
      (ip_ints[0] >= 0) &&
      (ip_ints[0] <= 255) &&
      (sscanf(ip_octets[1], "%d", &ip_ints[1]) == 1) &&
      (ip_ints[1] >= 0) &&
      (ip_ints[1] <= 255) &&
      (sscanf(ip_octets[2], "%d", &ip_ints[2]) == 1) &&
      (ip_ints[2] >= 0) &&
      (ip_ints[2] <= 255) &&
      (sscanf(ip_octets[3], "%d", &ip_ints[3]) == 1) &&
      (ip_ints[3] >= 0) &&
      (ip_ints[3] <= 255) &&
      (target_name != NULL) &&
      ((strlen_target_name = strlen(target_name)) > 0))
    {
    return_value = 0;

//printf("Looking for %s\n", target_ip); //FIXME
    /* This block looks for the dotted-quad IP address in the rDNS name. */
    return_value = search_ip(target_ip, strlen_target_ip, target_name, strlen_target_name);

    /*
     * This block looks for the zero-padded dotted-quad IP address in the rDNS
     * name.  For example, if the IP is 11.22.33.44, this block looks for
     * 011.022.033.044
     */
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%.3d.%.3d.%.3d.%.3d", ip_ints[0], ip_ints[1], ip_ints[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    /*
     * This block looks for the reversed dotted-quad IP address in the rDNS
     * name.  For example, if the IP is 11.22.33.44, this block looks for
     * 44.33.22.11.
     */
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s.%s.%s.%s", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    /*
     * This block looks for the slightly out-of-order dotted-quad IP address in the rDNS
     * name.  For example, if the IP is 11.22.33.44, this block looks for
     * 44.11.22.33
     */
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s.%s.%s.%s", ip_octets[3], ip_octets[0], ip_octets[1], ip_octets[2]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    /*
     * This block looks for the strangely out-of-order dotted-quad IP address in the rDNS
     * name.  For example, if the IP is 11.22.33.44, this block looks for
     * 33.22.11.44
     */
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s.%s.%s.%s", ip_octets[2], ip_octets[1], ip_octets[0], ip_octets[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    /*
     * This block looks for the strangely out-of-order dotted-quad IP address in the rDNS
     * name.  For example, if the IP is 11.22.33.44, this block looks for
     * 44.33.1122
     */
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s.%s.%s%s", ip_octets[3], ip_octets[2], ip_octets[0], ip_octets[1]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    /* This block looks for the strangely out-of-order dotted-quad IP address in the rDNS
     * name.  For example, if the IP is 11.22.33.44, this block looks for
     * 3344.11.22
     */
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s%s.%s.%s", ip_octets[2], ip_octets[3], ip_octets[0], ip_octets[1]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    /*
     * This block looks for the dotted-quad IP address with the last two octets encoded as a long integer in the rDNS
     * name.  For example, if the IP is 11.22.33.44, this block looks for
     * 11.22.8492
     */
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s.%s.%d", ip_octets[0], ip_octets[1], (ip_ints[2] * 256) + ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    /*
     * This block looks for the undotted-quad IP address in the rDNS name.
     * For example, if the IP is 11.22.33.44, this block looks for
     * 11223344.
     */
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%s%s%s%s", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    /*
     * This block looks for the undotted-, zero-padded-quad IP address in the
     * rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
     * 011022033044
     */
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%.3d%.3d%.3d%.3d", ip_ints[0], ip_ints[1], ip_ints[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    /*
     * This block looks for the undotted-, partially zero-padded-quad IP address in the
     * rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
     * 11022033044
     */
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%s%.3d%.3d%.3d", ip_octets[0], ip_ints[1], ip_ints[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    /*
     * This block looks for the undotted-, partially zero-padded-quad IP address in the
     * rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
     * 1122033044
     */
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%s%s%.3d%.3d", ip_octets[0], ip_octets[1], ip_ints[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    /*
     * This block looks for the undotted-, partially zero-padded-quad IP address in the
     * rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
     * 112233044
     */
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%s%s%s%.3d", ip_octets[0], ip_octets[1], ip_octets[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    /*
     * This block looks for the undotted-quad reversed IP address in the rDNS name.
     * For example, if the IP is 11.22.33.44, this block looks for
     * 44332211.
     */
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%s%s%s%s", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    /*
     * This block looks for the undotted-, zero-padded-quad reversed IP address in the
     * rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
     * 044033022011
     */
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%.3d%.3d%.3d%.3d", ip_ints[3], ip_ints[2], ip_ints[1], ip_ints[0]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    /*
     * This block looks for the IP address as an unsigned long integer.
     * For example, if the IP is 85.135.72.234, this block looks for
     * 3930621781.
     */
    if (!return_value &&
        inet_aton(target_ip, &tmp_addr))
      {
      snprintf(tmp_ip, MAX_IP, "%lu", (long unsigned int)tmp_addr.s_addr);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    /*
     * This block looks for the IP address as a sequence of hex bytes.
     * For example, if the IP is 85.135.72.234, this block looks for
     * 5080d7e3.
     */
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%.2x%.2x%.2x%.2x", ip_ints[0], ip_ints[1], ip_ints[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }
    }

  return(return_value);
  }

/*
 * Return value:
 *   DOES NOT RESOLVE: 0
 *   SUCCESS: 1
 */
int check_rdns_resolve(struct filter_settings *current_settings, char *target_hostname)
  {
  return(((target_hostname != NULL) && (target_hostname[0] != '\0')) ? dns_a(current_settings, target_hostname, NULL, 0) : 1);
  }

/*
 * Return value:
 *   NO COUNTRY CODE: 0
 *   COUNTRY CODE FOUND: 1
 */
int check_country_code(char *target_name, int strlen_target_name)
  {
  int return_value;
  int j;

  return_value = 0;

  if ((target_name != NULL) &&
      (strlen_target_name > 0))
    for (j = (strlen_target_name - 1); j >= 0; j--)
      if (target_name[j] == '.')
        {
        if ((strlen_target_name - j) == 3)
          return_value = 1;

        break;
        }

  return(return_value);
  }

/*
 * Return value:
 *   NOT FOUND: 0
 *   FOUND: matching line number within keywords_filename
 */
int check_rdns_keywords(struct filter_settings *current_settings, char *keywords_filename, char *target_name, int strlen_target_name)
  {
  int return_value;
  FILE *tmp_file;
  char keyword[MAX_FILE_BUF + 1];
  int line_num;
  int i;
  char *tld_start;
  int strlen_keyword;
  int strlen_subkeyword;
  char *name_ptr;
  int subkeyword_found;
  int at_least_one_not_found;

  return_value = 0;

  if ((target_name != NULL) &&
      (strlen_target_name > 0))
    {
    if ((tmp_file = fopen(keywords_filename, "r")) != NULL)
      {
      /* It's important to know where the domain name begins, since disqualifying mail.dhcp.com isn't fair. */
      tld_start = NULL;
      for (i = (strlen_target_name - 1); i >= 0; i--)
        if (target_name[i] == '.')
          {
          for (i--; i >= 0; i--)
            if (target_name[i] == '.')
              {
              tld_start = target_name + i + 1;
              break;
              }

          break;
          }

      if (tld_start != NULL)
        {
        line_num = 0;

        while (!feof(tmp_file) &&
               (line_num < MAX_FILE_LINES))
          {
          if ((fscanf(tmp_file, "%" STRINGIFY(MAX_FILE_BUF) "[^\r\n]", keyword) == 1) &&
              (keyword[0] != '#') &&
              ((strlen_keyword = strlen(keyword)) > 0))
            {
            for (i = 0; i < strlen_keyword; i++)
              if (keyword[i] == ' ')
                keyword[i] = '\0';

            i = 0;
            at_least_one_not_found = 0;
            while (i < strlen_keyword)
              {
              strlen_subkeyword = strlen(keyword + i);

              subkeyword_found = 0;
              name_ptr = target_name;
              while (((name_ptr - target_name) < strlen_target_name) &&
                     ((name_ptr = strstr(name_ptr, keyword + i)) != NULL) &&
                     ((keyword[i] == '.') ||
                      (name_ptr < tld_start)))
                {
                /*
                 * The preceding and following characters around the match must be
                 * non-alphanumeric so we don't reject 10.0.0.1.notdialup.example.com.
                 * This means 10.0.0.1.dialup3.example.com will get through but the
                 * alternative is worse.
                 *
                 * In pseudo-code: 
                 *   (((keyword is at start of name) ||
                 *     (first character in keyword is not a number or letter) ||
                 *     (character in name before keyword is not a number or letter)) &&
                 *    ((keyword is at end of name) ||
                 *     (last character in keyword is not a number or letter) ||
                 *     (character in name after keyword is not a number or letter)))
                 */
                if (((name_ptr == target_name) ||
                     !isalnum((int)(keyword + i)[0]) ||
                     !isalnum((int)(name_ptr - 1)[0])) &&
                    ((name_ptr == ((target_name + strlen_target_name) - strlen_subkeyword)) ||
                     !isalnum((int)(keyword + i + strlen_subkeyword - 1)[0]) ||
                     !isalnum((int)(name_ptr + strlen_subkeyword)[0])))
                  {
                  subkeyword_found = 1;
                  break;
                  }

                name_ptr += strlen_subkeyword;
                }

              if (!subkeyword_found)
                {
                at_least_one_not_found = 1;
                break;
                }

              i += strlen_subkeyword + 1;
              while ((i < strlen_keyword) &&
                     (keyword[i] == '\0'))
                i++;
              }

            if (!at_least_one_not_found)
              {
              return_value = line_num + 1;
              break;
              }
            }

          fscanf(tmp_file, "%*1[\r\n]");
          line_num++;
          }

        if (line_num == MAX_FILE_LINES)
          SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_FILE_TOO_LONG "%s", MAX_FILE_LINES, keywords_filename);

        fclose(tmp_file);
        }
      }
    else
      SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_OPEN_KEYWORDS "%s: %s", keywords_filename, strerror(errno));
    }

  return(return_value);
  }

/*
 * Return value:
 *   EARLYTALKER: 1
 *   NOT EARLYTALKER: 0
 */
int check_earlytalker(int delay)
  {
  int return_value;
  struct timeval tmp_time;
  fd_set listen_set;

  return_value = 0;

  tmp_time.tv_sec = delay;
  tmp_time.tv_usec = 0;

  FD_ZERO(&listen_set);
  FD_SET(STDIN_FD, &listen_set);
  if (select(STDIN_FD + 1, &listen_set, NULL, NULL, &tmp_time) == 1)
    return_value = 1;

  return(return_value);
  }

/*
 * Return value:
 *   FILTER_DECISION value
 */
int run_tests(struct filter_settings *current_settings, char *envp[])
  {
  int return_value;
  int i;
  int ip_in_name;
  int search_return;

  return_value = FILTER_DECISION_UNDECIDED;
  ip_in_name = -1;

  /* Check for rDNS name */
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      current_settings->check_rdns_exist &&
      (current_settings->strlen_server_name == 0))
    {
    current_settings->rejection = &rejection[REJECTION_RDNS_MISSING];
    return_value = FILTER_DECISION_DO_FILTER;
    }

  /* Check for IP and CC in name */
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      current_settings->check_ip_in_rdns_cc &&
      (current_settings->strlen_server_name > 0))
    {
    if (ip_in_name == -1)
      ip_in_name = is_ip_in_name(current_settings->server_ip, current_settings->server_name);

    if (ip_in_name &&
        check_country_code(current_settings->server_name, current_settings->strlen_server_name))
      {
      current_settings->rejection = &rejection[REJECTION_IP_IN_NAME_CC];
      return_value = FILTER_DECISION_DO_FILTER;
      }
    }

  /* Check rDNS whitelist */
  if ((return_value != FILTER_DECISION_DO_NOT_FILTER) &&
      (current_settings->whitelist_rdns_file != NULL) &&
      (current_settings->strlen_server_name > 0))
    for (i = 0; current_settings->whitelist_rdns_file[i] != NULL; i++)
      if (search_file(current_settings, current_settings->whitelist_rdns_file[i], current_settings->server_name, current_settings->strlen_server_name, '.', ".", '\0', NULL) > 0)
        {
        current_settings->rejection = NULL;
        return_value = FILTER_DECISION_DO_NOT_FILTER;
        break;
        }

  if ((return_value != FILTER_DECISION_DO_NOT_FILTER) &&
      (current_settings->whitelist_rdns_dir != NULL) &&
      (current_settings->strlen_server_name > 0))
    for (i = 0; current_settings->whitelist_rdns_dir[i] != NULL; i++)
      if (search_domain_directory(current_settings, current_settings->whitelist_rdns_dir[i], current_settings->server_name, current_settings->strlen_server_name) == 1)
        {
        current_settings->rejection = NULL;
        return_value = FILTER_DECISION_DO_NOT_FILTER;
        break;
        }

  /* Check rDNS blacklist */
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->blacklist_rdns_file != NULL) &&
      (current_settings->strlen_server_name > 0))
    for (i = 0; current_settings->blacklist_rdns_file[i] != NULL; i++)
      if ((search_return = search_file(current_settings, current_settings->blacklist_rdns_file[i], current_settings->server_name, current_settings->strlen_server_name, '.', ".", '\0', NULL)) > 0)
        {
        if (current_settings->log_level >= LOG_LEVEL_DEBUG)
          {
          memcpy(&current_settings->rejection_buf, &rejection[REJECTION_BLACKLIST_NAME], sizeof(struct rejection_data));
          snprintf(current_settings->short_reject_message_buf, MAX_BUF, "%s(%s:%d)", rejection[REJECTION_BLACKLIST_NAME].short_reject_message, current_settings->blacklist_rdns_file[i], search_return);
          current_settings->rejection_buf.short_reject_message = current_settings->short_reject_message_buf;
          current_settings->rejection = &current_settings->rejection_buf;
          }
        else
          current_settings->rejection = &rejection[REJECTION_BLACKLIST_NAME];

        return_value = FILTER_DECISION_DO_FILTER;

        break;
        }

  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->blacklist_rdns_dir != NULL) &&
      (current_settings->strlen_server_name > 0))
    for (i = 0; current_settings->blacklist_rdns_dir[i] != NULL; i++)
      if (search_domain_directory(current_settings, current_settings->blacklist_rdns_dir[i], current_settings->server_name, current_settings->strlen_server_name) == 1)
        {
        current_settings->rejection = &rejection[REJECTION_BLACKLIST_NAME];
        return_value = FILTER_DECISION_DO_FILTER;
        break;
        }

  /* Check IP whitelist */
  if (current_settings->whitelist_ip_file != NULL)
    for (i = 0; current_settings->whitelist_ip_file[i] != NULL; i++)
      if (search_tcprules_file(current_settings, NULL, 0, current_settings->whitelist_ip_file[i], envp, current_settings->server_ip, current_settings->server_name, current_settings->strlen_server_name) > 0)
        {
        current_settings->rejection = NULL;
        return_value = FILTER_DECISION_DO_NOT_FILTER;
        break;
        }

  /* Check IP blacklist */
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->blacklist_ip_file != NULL))
    for (i = 0; current_settings->blacklist_ip_file[i] != NULL; i++)
      if ((search_return = search_tcprules_file(current_settings, NULL, 0, current_settings->blacklist_ip_file[i], envp, current_settings->server_ip, current_settings->server_name, current_settings->strlen_server_name)) > 0)
        {
        if (current_settings->log_level >= LOG_LEVEL_DEBUG)
          {
          memcpy(&current_settings->rejection_buf, &rejection[REJECTION_BLACKLIST_IP], sizeof(struct rejection_data));
          snprintf(current_settings->short_reject_message_buf, MAX_BUF, "%s(%s:%d)", rejection[REJECTION_BLACKLIST_IP].short_reject_message, current_settings->blacklist_ip_file[i], search_return);
          current_settings->rejection_buf.short_reject_message = current_settings->short_reject_message_buf;
          current_settings->rejection = &current_settings->rejection_buf;
          }
        else
          current_settings->rejection = &rejection[REJECTION_BLACKLIST_IP];

        return_value = FILTER_DECISION_DO_FILTER;

        break;
        }

  /* Check for IP and keywords in name */
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->blacklist_rdns_keyword_file != NULL) &&
      (current_settings->strlen_server_name > 0))
    {
    if (ip_in_name == -1)
      ip_in_name = is_ip_in_name(current_settings->server_ip, current_settings->server_name);

    if (ip_in_name)
      for (i = 0; current_settings->blacklist_rdns_keyword_file[i] != NULL; i++)
        if ((search_return = check_rdns_keywords(current_settings, current_settings->blacklist_rdns_keyword_file[i], current_settings->server_name, current_settings->strlen_server_name)))
          {
          if (current_settings->log_level >= LOG_LEVEL_DEBUG)
            {
            memcpy(&current_settings->rejection_buf, &rejection[REJECTION_IP_IN_NAME], sizeof(struct rejection_data));
            snprintf(current_settings->short_reject_message_buf, MAX_BUF, "%s(%s:%d)", rejection[REJECTION_IP_IN_NAME].short_reject_message, current_settings->blacklist_rdns_keyword_file[i], search_return);
            current_settings->rejection_buf.short_reject_message = current_settings->short_reject_message_buf;
            current_settings->rejection = &current_settings->rejection_buf;
            }
          else
            current_settings->rejection = &rejection[REJECTION_IP_IN_NAME];

          return_value = FILTER_DECISION_DO_FILTER;

          break;
          }
    }

  /* Check for rDNS resolution */
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      current_settings->check_rdns_resolve &&
      (current_settings->strlen_server_name > 0) &&
      (((strncasecmp(current_settings->server_name, LOCALHOST_NAME, STRLEN(LOCALHOST_NAME)) != 0) &&
        !check_rdns_resolve(current_settings, current_settings->server_name)) ||
       ((strncasecmp(current_settings->server_name, LOCALHOST_NAME, STRLEN(LOCALHOST_NAME)) == 0) &&
        (strncasecmp(current_settings->server_ip, LOCALHOST_IP, STRLEN(LOCALHOST_IP)) != 0))))
    {
    current_settings->rejection = &rejection[REJECTION_RDNS_RESOLVE];
    return_value = FILTER_DECISION_DO_FILTER;
    }

  /* Check DNS RWL */
  if ((return_value != FILTER_DECISION_DO_NOT_FILTER) &&
      (current_settings->dnsrwl_fqdn != NULL) &&
      (strcmp(current_settings->server_ip, LOCALHOST_IP) != 0))
    for (i = 0; current_settings->dnsrwl_fqdn[i] != NULL; i++)
      if (check_dnsrbl(current_settings, current_settings->server_ip, current_settings->dnsrwl_fqdn[i], NULL))
        {
        return_value = FILTER_DECISION_DO_NOT_FILTER;
        current_settings->rejection = NULL;
        break;
        }

  /* Check RHSWL */
  if ((return_value != FILTER_DECISION_DO_NOT_FILTER) &&
      (current_settings->rhswl_fqdn != NULL) &&
      (strcmp(current_settings->server_ip, LOCALHOST_IP) != 0) &&
      (current_settings->strlen_server_name > 0))
    for (i = 0; current_settings->rhswl_fqdn[i] != NULL; i++)
      if (check_rhsbl(current_settings, current_settings->server_name, current_settings->rhswl_fqdn[i], NULL))
        {
        return_value = FILTER_DECISION_DO_NOT_FILTER;
        current_settings->rejection = NULL;
        break;
        }

  /* Check DNS RBL */
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->dnsrbl_fqdn != NULL) &&
      (strcmp(current_settings->server_ip, LOCALHOST_IP) != 0))
    for (i = 0; current_settings->dnsrbl_fqdn[i] != NULL; i++)
      if (check_dnsrbl(current_settings, current_settings->server_ip, current_settings->dnsrbl_fqdn[i], &rejection[REJECTION_RBL]))
        {
        return_value = FILTER_DECISION_DO_FILTER;
        break;
        }

  /* Check RHSBL */
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->rhsbl_fqdn != NULL) &&
      (strcmp(current_settings->server_ip, LOCALHOST_IP) != 0) &&
      (current_settings->strlen_server_name > 0))
    for (i = 0; current_settings->rhsbl_fqdn[i] != NULL; i++)
      if (check_rhsbl(current_settings, current_settings->server_name, current_settings->rhsbl_fqdn[i], &rejection[REJECTION_RHSBL]))
        {
        return_value = FILTER_DECISION_DO_FILTER;
        break;
        }

  /* Check for earlytalkers */
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->check_earlytalker > 0) &&
      check_earlytalker(current_settings->check_earlytalker))
    {
    current_settings->rejection = &rejection[REJECTION_EARLYTALKER];
    return_value = FILTER_DECISION_DO_FILTER;
    }

  if (return_value == FILTER_DECISION_UNDECIDED)
    return_value = current_settings->filter_action;

  return(return_value);
  }

/*
 * Return value:
 *   FILTER_DECISION value
 */
int process_access(struct filter_settings *current_settings, char *envp[], char ***destination_envp)
  {
  int return_value;
  int i;
  int j;
  int k;
  char command_text[MAX_BUF + 1];
  char *current_ptr;
  char *value_start_ptr;
  char *value_end_ptr;
  int strlen_envp;
  int len_envp;
  char tmp_delimiter;
  void *tmp_ptr;

  return_value = FILTER_DECISION_UNDECIDED;
  command_text[0] = '\0';

  if (current_settings->access_list_file != NULL)
    for (i = 0; current_settings->access_list_file[i] != NULL; i++)
      if (search_tcprules_file(current_settings, command_text, MAX_BUF, current_settings->access_list_file[i], envp, current_settings->server_ip, current_settings->server_name, current_settings->strlen_server_name) >= 0)
        {
        current_settings->allow_relay = 0;

        if ((command_text[0] == '\0') ||
            ((strncasecmp(command_text, TCPRULES_ALLOW, STRLEN(TCPRULES_ALLOW)) == 0) &&
             ((command_text[STRLEN(TCPRULES_ALLOW)] == TCPRULES_DELIMITER) ||
              (command_text[STRLEN(TCPRULES_ALLOW)] == '\0'))))
          {
          return_value = FILTER_DECISION_DO_FILTER;

          if ((*destination_envp == NULL) ||
              (*destination_envp == envp))
            {
            if (envp != NULL)
              {
              for (len_envp = 0; envp[len_envp] != NULL; len_envp++);
              if ((*destination_envp = malloc(sizeof(char *) * (len_envp + 1))) != NULL)
                {
                (*destination_envp)[0] = NULL;
                for (j = 0; j < len_envp; j++)
                  {
                  strlen_envp = strlen(envp[j]);

                  if (strncasecmp(envp[j], ENVIRONMENT_ALLOW_RELAY, STRLEN(ENVIRONMENT_ALLOW_RELAY)) == 0)
                    current_settings->allow_relay = 1;

                  if (((*destination_envp)[j] = malloc(sizeof(char) * (strlen_envp + 1))) != NULL)
                    {
                    strncpy((*destination_envp)[j], envp[j], strlen_envp);
                    (*destination_envp)[j][strlen_envp] = '\0';
                    (*destination_envp)[j + 1] = NULL;
                    }
                  else
                    {
                    for (k = 0; k < j; k++)
                      free((*destination_envp)[k]);
                    free(*destination_envp);
                    *destination_envp = NULL;

                    SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_REALLOC, (unsigned long)(sizeof(char) * (strlen_envp + 1)));
                    return_value = FILTER_DECISION_ERROR;
                    break;
                    }
                  }
                }
              else
                {
                SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_REALLOC, (unsigned long)(sizeof(char *) * (len_envp + 1)));
                return_value = FILTER_DECISION_ERROR;
                break;
                }
              }
            else
              {
              *destination_envp = NULL;
              len_envp = 0;
              }
            }
          else
            for (len_envp = 0; (*destination_envp)[len_envp] != NULL; len_envp++);

        if ((strncasecmp(command_text, TCPRULES_ALLOW, STRLEN(TCPRULES_ALLOW)) == 0) &&
            (command_text[STRLEN(TCPRULES_ALLOW)] == TCPRULES_DELIMITER))
          {
          current_ptr = command_text + STRLEN(TCPRULES_ALLOW) + 1;
          while (((current_ptr - command_text) < MAX_BUF) &&
                 (current_ptr[0] != '\0'))
            {
            for (value_start_ptr = current_ptr; (((value_start_ptr - command_text) < (MAX_BUF - 3)) && (value_start_ptr[0] != '\0') && (value_start_ptr[0] != ENVIRONMENT_DELIMITER)); value_start_ptr++);
            if (value_start_ptr[0] == ENVIRONMENT_DELIMITER)
              {
              for (j = 0; j < len_envp; j++)
                if (strncasecmp((*destination_envp)[j], current_ptr, value_start_ptr - current_ptr) == 0)
                  break;

              tmp_delimiter = value_start_ptr[1];
              value_start_ptr += 2;
              for (value_end_ptr = value_start_ptr; (((value_end_ptr - command_text) < (MAX_BUF - 1)) && (value_end_ptr[0] != '\0') && (value_end_ptr[0] != tmp_delimiter)); value_end_ptr++);
              if (j == len_envp)
                {
                if ((tmp_ptr = realloc(*destination_envp, sizeof(char *) * (len_envp + 2))) != NULL)
                  {
                  *destination_envp = tmp_ptr;
                  (*destination_envp)[len_envp] = NULL;
                  (*destination_envp)[len_envp + 1] = NULL;
                  len_envp++;
                  }
                else
                  {
                  SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_REALLOC, (unsigned long)(sizeof(char *) * (len_envp + 2)));
                  return_value = FILTER_DECISION_ERROR;
                  break;
                  }
                }

              if ((tmp_ptr = realloc((*destination_envp)[j], sizeof(char) * (value_end_ptr - current_ptr))) != NULL)
                (*destination_envp)[j] = tmp_ptr;
              else
                {
                SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_REALLOC, (unsigned long)(sizeof(char) * ((value_end_ptr - value_start_ptr) + 1)));
                return_value = FILTER_DECISION_ERROR;
                break;
                }

              if (strncasecmp(current_ptr, ENVIRONMENT_ALLOW_RELAY, STRLEN(ENVIRONMENT_ALLOW_RELAY)) == 0)
                current_settings->allow_relay = 1;

              strncpy((*destination_envp)[j], current_ptr, (value_start_ptr - current_ptr) - 1);
              strncpy((*destination_envp)[j] + ((value_start_ptr - current_ptr) - 1), value_start_ptr, value_end_ptr - value_start_ptr);
              (*destination_envp)[j][(value_end_ptr - current_ptr) - 1] = '\0';
              current_ptr = value_end_ptr + 2;
              }
            else
              current_ptr = value_start_ptr;
            }
          }

        if (!current_settings->allow_relay &&
            (current_settings->local_domains_file != NULL))
          {
          if ((tmp_ptr = realloc(*destination_envp, sizeof(char *) * (len_envp + 2))) != NULL)
            {
            *destination_envp = tmp_ptr;
            (*destination_envp)[len_envp] = NULL;
            (*destination_envp)[len_envp + 1] = NULL;
            len_envp++;

            if ((tmp_ptr = realloc((*destination_envp)[len_envp - 1], sizeof(char) * (STRLEN(ENVIRONMENT_ALLOW_RELAY) + 2))) != NULL)
              {
              (*destination_envp)[len_envp - 1] = tmp_ptr;
              strncpy((*destination_envp)[len_envp - 1], ENVIRONMENT_ALLOW_RELAY, STRLEN(ENVIRONMENT_ALLOW_RELAY));
              (*destination_envp)[len_envp - 1][STRLEN(ENVIRONMENT_ALLOW_RELAY)] = ENVIRONMENT_DELIMITER;
              (*destination_envp)[len_envp - 1][STRLEN(ENVIRONMENT_ALLOW_RELAY) + 1] = '\0';
              }
            else
              {
              SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_REALLOC, (unsigned long)(sizeof(char) * (STRLEN(ENVIRONMENT_ALLOW_RELAY) + 2)));
              return_value = FILTER_DECISION_ERROR;
              break;
              }
            }
          else
            {
            SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_REALLOC, (unsigned long)(sizeof(char *) * (len_envp + 2)));
            return_value = FILTER_DECISION_ERROR;
            break;
            }
          }

        break;
        }
      else if ((strncasecmp(command_text, TCPRULES_DENY, STRLEN(TCPRULES_DENY)) == 0) &&
               ((command_text[STRLEN(TCPRULES_DENY)] == TCPRULES_DELIMITER) ||
                (command_text[STRLEN(TCPRULES_DENY)] == '\0')))
        {
        current_settings->rejection = &rejection[REJECTION_ACCESS_DENIED];
        return_value = FILTER_DECISION_DO_FILTER;
        break;
        }
      }

  return(return_value);
  }

/*
 * Return value:
 *   0
 */
int do_spamdyke(struct filter_settings *current_settings, int argc, char *argv[], char *envp[])
  {
  int output_pipe[2];
  int input_pipe[2];
  int child_pid;
  int test_result;
  char **new_envp;

  srandom(time(NULL));
  signal(SIGPIPE, SIG_IGN);

  new_envp = envp;

  if ((((current_settings->filter_action = process_command_line(current_settings, argc, argv, envp)) == FILTER_DECISION_UNDECIDED) ||
       (current_settings->filter_action == FILTER_DECISION_DO_FILTER)) &&
      (((test_result = process_access(current_settings, envp, &new_envp)) == FILTER_DECISION_UNDECIDED) ||
       (test_result == FILTER_DECISION_DO_FILTER)))
    {
    if (current_settings->filter_action == FILTER_DECISION_UNDECIDED)
      current_settings->filter_action = test_result;

    if ((test_result = run_tests(current_settings, new_envp)) != FILTER_DECISION_UNDECIDED)
      current_settings->filter_action = test_result;
    }

#ifdef HAVE_LIBSSL
  if ((current_settings->tls_certificate_file != NULL) &&
      !tls_init(current_settings))
    {
    current_settings->tls_certificate_file = NULL;
    current_settings->tls_privatekey_file = NULL;
    current_settings->tls_privatekey_password = NULL;
    }
#endif /* HAVE_LIBSSL */

  output_pipe[0] = -1;
  output_pipe[1] = -1;
  input_pipe[0] = -1;
  input_pipe[1] = -1;

  if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
      (current_settings->filter_grace <= FILTER_GRACE_NONE) &&
      (current_settings->rejection != NULL))
    {
    output_write_rejection(current_settings, current_settings->rejection, STDOUT_FD, REJECT_SEVERITY_HELO);

    middleman(&input_pipe[0], &output_pipe[1], '\n', current_settings, -1, new_envp);
    }
  else
    {
    if (pipe(output_pipe) != -1)
      {
      if (pipe(input_pipe) != -1)
        {
        if ((child_pid = fork()) > 0)
          /* parent */
          {
          close(output_pipe[0]);
          close(input_pipe[1]);

          middleman(&input_pipe[0], &output_pipe[1], '\n', current_settings, child_pid, new_envp);

          if (output_pipe[1] != -1)
            close(output_pipe[1]);
          if (input_pipe[0] != -1)
            close(input_pipe[0]);

          waitpid(child_pid, NULL, WNOHANG);
          }
        else if (child_pid == 0)
          /* child */
          {
          close(output_pipe[1]);
          close(input_pipe[0]);

          if ((dup2(output_pipe[0], STDIN_FD) != -1) &&
              (dup2(input_pipe[1], STDOUT_FD) != -1))
            {
            signal(SIGPIPE, SIG_DFL);

            if (!exec_path(current_settings->child_argv[0], current_settings->child_argv, new_envp))
              {
              SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_EXEC "%s: %s", current_settings->child_argv[0], strerror(errno));
              close(output_pipe[0]);
              close(input_pipe[1]);
              }
            }
          else
            {
            SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

            close(output_pipe[0]);
            close(input_pipe[1]);

            current_settings->filter_action = FILTER_DECISION_FORK_ERROR;
            }
          }
        else
          {
          SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_FORK "%s", strerror(errno));

          close(output_pipe[0]);
          close(output_pipe[1]);
          close(input_pipe[0]);
          close(input_pipe[1]);

          current_settings->filter_action = FILTER_DECISION_FORK_ERROR;
          }
        }
      else
        {
        SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_PIPE "%s", strerror(errno));

        close(output_pipe[0]);
        close(output_pipe[1]);

        current_settings->filter_action = FILTER_DECISION_FORK_ERROR;
        }
      }
    else
      {
      SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_PIPE "%s", strerror(errno));
      current_settings->filter_action = FILTER_DECISION_FORK_ERROR;
      }
    }

  if ((current_settings->filter_action == FILTER_DECISION_FORK_ERROR) &&
      !exec_path(current_settings->child_argv[0], current_settings->child_argv, new_envp))
    SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_EXEC "%s: %s", current_settings->child_argv[0], strerror(errno));

#ifdef HAVE_LIBSSL
  if (tls_end(current_settings, STDIN_FD))
    output_writeln(current_settings, LOG_ACTION_TLS_END, -1, NULL, -1);
#endif /* HAVE_LIBSSL */

  dns_initialize(current_settings, 1);
  output_writeln(NULL, 0, -1, NULL, 0);
  SPAMDYKE_LOG_NONE(current_settings, NULL);

  return(0);
  }

int main(int argc, char *argv[], char *envp[])
  {
  return(prepare_settings(argc, argv, envp, &do_spamdyke));
  }
